//##################################################
//# $Id: Main.cpp 19 2005-03-01 15:15:06Z peciva $

/*****************************************************************************\
 *
 * Main.cpp
 *
 * Test application for Infinite Terrain Engine
 *
 * Authors: Martin Havlek (xhavli15 AT stud.fit.vutbr.cz)
 *          PCJohn (peciva AT fit.vutbr.cz)
 * Contributors: 
 *
 * ----------------------------------------------------------------------------
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, authors will kindly welcome
 * if you give them credit and keep their names with their source code.
 *
\*****************************************************************************/


#ifdef _WIN32
#  include <Inventor/Win/SoWin.h>
#  include <Inventor/Win/SoWinRenderArea.h>
#  pragma comment(linker, "/entry:\"mainCRTStartup\"")
#  pragma comment(linker, "/subsystem:\"console\"")
#else
#  include <Inventor/Qt/SoQt.h>
#  include <Inventor/Qt/SoQtRenderArea.h>
#endif

#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoDirectionalLight.h>
#include <Inventor/nodes/SoPerspectiveCamera.h>
#include <Inventor/nodes/SoEventCallback.h>
#include <Inventor/nodes/SoFile.h>

#include <Inventor/events/SoEvent.h>

#include <Inventor/sensors/SoOneShotSensor.h>

#include <cstdio>  // printf
#include <cstdlib> // putenv, exit
#include <cmath>   // floor
#include <ctime>   // time

#include "TeEngine.h"
#include "objects/TeEngineExplorer.h"


#define APP_TITLE "TerrainEngine"
#define APP_LAST_UPDATE "2005-05-08"
#define VIEWER_BGCOLOR SbColor(0.33f, 0.74f, 0.93f)

// ??? TEMPORARY:
// use "app-object" instead and store all important values in it ???
static TeEngine *engine = NULL;


//----------------------------------------------------------------------------
// Time-tick callback
static void timeTickCB(void *data, SoSensor *sensor)
{
  static SbTime last = SbTime::zero();

  SbTime delta;
  SbTime current = SbTime::getTimeOfDay();

  if (last == SbTime::zero()) delta = SbTime::zero();
  else delta = current - last;

  last = current;

  // call engine's timeTick
  engine->timeTick(delta);

  sensor->schedule();
};


//-----------------------------------------------------------------------------
// Event callback
static void eventCBFunc(void *userData, SoEventCallback *eventCB)
{
  const SoEvent *event = eventCB->getEvent();
  engine->getActiveObject()->handleEvent(event);
  eventCB->setHandled();
}


//----------------------------------------------------------------------------
// main
int main(int argc, char *argv[])
{
  // SoWin & Coin3D libraries init
  #ifdef _WIN32
    HWND window = SoWin::init(argv[0]);
  #else
    QWidget *window = SoQt::init(argv[0]);
  #endif
  if (window == NULL) exit(1);

  // random num. generator init
  srand((unsigned int)time(NULL));

  // Directory with models and textures
  SoInput::addDirectoryFirst("data/textures/");
  SoInput::addDirectoryFirst("data/meshes/");
  SoInput::addDirectoryFirst("../data/textures/");
  SoInput::addDirectoryFirst("../data/meshes/");

  // scene root
  SoSeparator *root = new SoSeparator;
  root->ref();

  // scene separator
  SoSeparator *sceneRoot = new SoSeparator;
  root->addChild(sceneRoot);

  // scene camera
  SoPerspectiveCamera *sceneCamera = new SoPerspectiveCamera;
  sceneCamera->nearDistance = 0.1f;
  sceneCamera->farDistance = 250.0f;
  sceneCamera->orientation.setValue( SbRotation(SbVec3f(1.0f, 0.0f, 0.0f), (float)M_PI/180.0f*80.0f) );
  sceneRoot->addChild(sceneCamera);

  // sunlight
  SoDirectionalLight *sunLight = new SoDirectionalLight;
  sunLight->direction.setValue(0.75f, 0.5f, -1.0f);
  sunLight->intensity.setValue(1.0);
  sceneRoot->addChild(sunLight);

  // terrain engine
  engine = new TeEngine(SbVec2s(32,65), SbVec2f(128.0f,256.0f));
  engine->faultFormation.maxDelta = 5.0f;
  engine->setSceneRoot(sceneRoot);

  // simple engine object, it will fly along +y axis at constant speed
  TeObject *object = new TeObject;
  object->setModel("Bomber.wrl", TeObject::RENDER);
  object->setModel("Bomber-C.wrl", TeObject::COLLISION);
  object->position = SbVec3f(-60.0f, -30.0f, 210.0f);
  object->speed = SbVec3f(0.0f, 10.0f, 0.0f);
  engine->addObject(object);

  // engine explorer - active object
  // !!! ACTIVE OBJECT SHOULD BE THE LAST ONE IN THE LIST !!!
  // !!! It will cause remove/add blink if it is not...   !!!
  TeEngineExplorer *explorer = new TeEngineExplorer;
  explorer->setModel("Bomber.wrl", TeObject::RENDER);
  explorer->setModel("Bomber-C.wrl", TeObject::COLLISION);
  explorer->position = SbVec3f(0.0f, -30.0f, 210.0f);
  engine->addObject(explorer);
  engine->setActiveObject(explorer);

  // HUD separator
  SoSeparator *HUDRoot = new SoSeparator;
  root->addChild(HUDRoot);

  // HUD camera
  SoPerspectiveCamera *HUDCamera = new SoPerspectiveCamera;
  HUDCamera->nearDistance = 0.1f;
  HUDCamera->farDistance = 10.0f;
  HUDRoot->addChild(HUDCamera);

  // engine init
  engine->setHUDRoot(HUDRoot);
  engine->setSceneCamera(sceneCamera);
  engine->setHUDCamera(HUDCamera);
  engine->initialize();

  // render area
  #ifdef _WIN32
    SoWinRenderArea *ra = new SoWinRenderArea(window);
  #else
    SoQtRenderArea *ra = new SoQtRenderArea(window);
  #endif
  ra->setBackgroundColor( VIEWER_BGCOLOR );
  ra->setSceneGraph(root);
  ra->setTitle( APP_TITLE );
  ra->show();

  // event callback
  SoEventCallback *eventCallback = new SoEventCallback;
  eventCallback->addEventCallback(SoEvent::getClassTypeId(), eventCBFunc, NULL);
  root->addChild(eventCallback);

  // timeTick sensor
  SoOneShotSensor *timeTickSensor = new SoOneShotSensor(timeTickCB, NULL);
  timeTickSensor->schedule();

  // console echo
  printf("%s (%s)\n\nRunning...\n\n", APP_TITLE, APP_LAST_UPDATE);
  printf("Use arrows to change object's speed vector.\nUse 8, 6, 2, 4 on NumPad, PdUp and PgDown to change camera's distance.\n\n");

  ra->setFullScreen(TRUE);

  // show window and run mainLoop
  #ifdef _WIN32
    SoWin::show(window);
    SoWin::mainLoop();
  #else
    SoQt::show(window);
    SoQt::mainLoop();
  #endif

  // free memory
  delete ra;
  delete explorer;
  root->unref();

  // console echo
  printf("\nDone...\n");

  return 0;
}
