//##################################################
//# $Id: TePatch.cpp 20 2005-03-14 13:15:11Z peciva $

/*****************************************************************************\
 *
 * TePatch.cpp
 *
 * TePatch - terrain patch class
 *
 * Author: PCJohn (peciva AT fit.vutbr.cz)
 * Contributors: Martin Havlek (xhavli15 AT stud.fit.vutbr.cz)
 *
 * ----------------------------------------------------------------------------
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, authors will kindly welcome
 * if you give them credit and keep their names with their source code.
 *
\*****************************************************************************/


#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoIndexedTriangleStripSet.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoMaterialBinding.h>
#include <Inventor/nodes/SoNormal.h>
#include <Inventor/nodes/SoNormalBinding.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoTranslation.h>
#include <Inventor/errors/SoDebugError.h>

#include <ctime>

#include "TePatch.h"
#include "TeHeightMap.h"
#include "TeEngine.h"


//-----------------------------------------------------------------------------
/**
 * Simple default constructor.
 *
 * It does not specify the engine that manages this patch and therefore is
 * private and shoul be \b never called. Documented only to explain the
 * reason.
 */
TePatch::TePatch()
{
  assert(0 && "Default constructor of TePatch is private and should be never called.");
}


//-----------------------------------------------------------------------------
/**
 * Constructor.
 *
 * \param aengine Engine in which patch hierarchy this patch belongs.
 */
TePatch::TePatch(TeEngine *aengine) : hmap(NULL), engine(aengine), parent(NULL), graph(NULL)
{
  assert(engine && "Engine parameter can not be NULL.");
  int i;
  for (i=0;i<8;i++)  seamMap[i] = NULL;
  for (i=0;i<4;i++)  buildMap[i] = NULL;
  for (i=0;i<4;i++)  detail[i] = NULL;
  for (i=0;i<8;i++)  neighbour[i] = NULL;
  for (i=0;i<8;i++)  seamGraph[i] = NULL;
}


//-----------------------------------------------------------------------------
/**
 * Destructor.
 *
 * Releases memory.
 *
 * \todo Cancel scheduled tasks at engine. ???
 */
TePatch::~TePatch()
{
  int i;

  // TODO: cancel scheduled tasks at engine

  // delete more detail patches
  for (i=0; i<4; i++)  delete detail[i];

  // delete buildMaps
  if (!neighbour[SOUTH_WEST] && !neighbour[SOUTH] && !neighbour[WEST])  delete buildMap[SOUTH_WEST];
  if (!neighbour[SOUTH_EAST] && !neighbour[SOUTH] && !neighbour[EAST])  delete buildMap[SOUTH_EAST];
  if (!neighbour[NORTH_WEST] && !neighbour[NORTH] && !neighbour[WEST])  delete buildMap[NORTH_WEST];
  if (!neighbour[NORTH_EAST] && !neighbour[NORTH] && !neighbour[EAST])  delete buildMap[NORTH_EAST];

  // release heightmap
  delete hmap;

  // release seam height maps
  for (i=0; i<8; i++)  delete seamMap[i];

  // release scene graphs
  if (graph)  graph->unref();
  for (i=0; i<8; i++)
    if (seamGraph[i])  seamGraph[i]->unref();

  // remove self from neighbours references
  for (i=0; i<8; i++)
    if (neighbour[i]) {
      neighbour[i]->neighbour[getOppositeDirection(i)] = NULL;
    }

  // remove parent reference
  if (parent) { // this is always true if we are not root node
    for (i=0; i<4; i++) {
      if (parent->detail[i] == this) {
        parent->detail[i] = NULL;
        break;
      }
    }
    assert(i<4 && "This patch was not found in the parent patch!");
  }
  else
    if (engine->getRootPatch() == this)
      engine->onRootDestroy();
}


//-----------------------------------------------------------------------------
/**
 * Private attribute accessor.
 *
 * \return Const pointer to the patch heightmap.
 *
 * \sa hmap, setHMap()
 */
const TeHeightMap* TePatch::getHMap() const
{
  return hmap;
}


//-----------------------------------------------------------------------------
/**
 * Private attribute accessor.
 *
 * \return Pointer to the patch heightmap.
 *
 * \sa hmap, setHMap()
 */
TeHeightMap* TePatch::getHMap()
{
  return hmap;
}


//-----------------------------------------------------------------------------
/**
 * Sets private hmap attribute.
 *
 * This function checks if the resolution of the heightmap \a hmap is the same
 * as the one set in the engine. If it is not so, nothing happens and the
 * warning is post. If there is already some heightmap set, it is deleted.
 *
 * \param hmap Pointer to the heightmap.
 *
 * \todo When the heightmap is being deleted, we probably should unref all
 *       scene graphs and delete the seam maps too.
 *
 * \sa hmap, getHMap()
 */
void TePatch::setHMap(TeHeightMap *hmap)
{
  if (hmap->getResolution() != engine->getHMapResolution()) {
    SoDebugError::post("TePatch:setHMap",
                       "HMap resolution must be the same as the one set in TeEngine constructor,\n"
                       "that usually defaults to 65x65. The problem is handled by ignoring of this call.");
    return;
  }

  if (this->hmap)
    delete this->hmap;

  this->hmap = hmap;
}


//-----------------------------------------------------------------------------
/**
 * Private attribute accessor.
 *
 * \param which Seam map selector.
 * \return Const pointer to the specified seam map.
 *
 * \sa seamMap, setSeamMap()
 */
const TeHeightMap* TePatch::getSeamMap(const Direction which) const
{
  assert(which>=0 && which<8 && "Wrong which parameter.");
  return seamMap[which];
}


//-----------------------------------------------------------------------------
/**
 * Private attribute accessor.
 *
 * \param which Seam map selector.
 * \return Pointer to the specified seam map.
 *
 * \sa seamMap, setSeamMap()
 */
TeHeightMap* TePatch::getSeamMap(const Direction which)
{
  assert(which>=0 && which<8 && "Wrong which parameter.");
  return seamMap[which];
}


//-----------------------------------------------------------------------------
/**
 * Sets private seam map attribute.
 *
 * This function checks if the resolution of the heightmap \a map corresponds
 * with the one set in the engine and the selected seam. If it is not so,
 * nothing happens and the warning is post.
 *
 * \param which Seam map selector.
 * \param map Pointer to the heightmap.
 *
 * \sa seamMap, getSeamMap()
 *
 * \todo If there is already some seam map set, it should be deleted. It is
 *       similar to the TePatch::setHMap() function.
 */
void TePatch::setSeamMap(const Direction which, TeHeightMap *map)
{
  assert(which>=0 && which<8 && "Wrong which parameter.");

  int rx,ry;
  map->getResolution(rx,ry);
  if (which < 4) {
    if (rx != 5 || ry != 5)  goto bad_res;
  }
  else {
    const SbVec2s &res = engine->getHMapResolution();
    if (which == WEST || which == EAST) {
      if (rx != 5 || ry != res[1])  goto bad_res;
    } else {
      if (ry != 5 || rx != res[0])  goto bad_res;
    }
  }


  seamMap[which] = map;
  return;

bad_res:
  SoDebugError::post("TePatch:setSeamMap",
                     "Wrong seam height maps resolution. It must be [5,TeEngine::\n"
                     "getHMapResolution()[1], or [TeEngine::getHMapResolution()[0],5],\n"
                     "or [5,5], depending of which parameter. Setting of the height map is ignored.");
}


//-----------------------------------------------------------------------------
/**
 * Private attribute accessor.
 *
 * \return Pointer to the attached engine.
 *
 * \sa engine, setEngine()
 */
TeEngine* TePatch::getEngine() const
{
  return engine;
}


//-----------------------------------------------------------------------------
/**
 * Should set the private engine attribute.
 *
 * This function is not implemented yet, it looks like we don't need it...
 *
 * \sa engine, getEngine()
 *
 * \todo Not implemented yet. Consider removing it completely, because its
 *       work is done in constructor and changing attached engines at
 *       run-time will be strange hack.
 */
void TePatch::setEngine()
{
  assert(0 && "Not implemented yet.");
}


//-----------------------------------------------------------------------------
/**
 * Private attribute accessor.
 *
 * \return Rectangle area occupied by the scene graph.
 *
 * \sa area
 */
const SbBox2f& TePatch::getArea() const
{
  return area;
}


//-----------------------------------------------------------------------------
/**
 * Gets level of the patch in the hierarchy.
 *
 * \return Patch level in the hierarchy.
 *
 * \todo More detailed documentation.
 */
int TePatch::getLevel() const
{
  if (parent)
    return parent->getLevel() - 1;
  else
    if (engine->getRootPatch() == this)  return engine->getRootLevel();
    else {
      SoDebugError::post("TePatch::getLevel",
                         "It is forbidden to call getLevel() until TePatch is made part of the\n"
                         "TePatch hierarchy headed by an TeEngine.");
      return 0;
    }
}


//-----------------------------------------------------------------------------
/**
 * Private attribute accessor.
 *
 * \return Parent patch in the hierarchy.
 *
 * \sa parent
 *
 * \todo More detailed documentation.
 */
TePatch* TePatch::getParent() const
{
  return parent;
}


//-----------------------------------------------------------------------------
/**
 * Documentation not available.
 *
 * \todo More detailed documentation.
 */
TePatch* TePatch::getPatchHierarch(const SbVec2f &pos, const int level, PatchMissedPolicy policy)
{
  return getPatchHierarchRelative(pos, level - this->getLevel(), policy);
}


//-----------------------------------------------------------------------------
/**
 * Documentation not available.
 *
 * \todo More detailed documentation.
 */
TePatch* TePatch::getPatchHierarchRelative(const SbVec2f &pos, const int relativeLevel, PatchMissedPolicy policy)
{
  // policy works like this:
  //   - DONT_CREATE  - return NULL
  //   - CREATE       - create all necesary patches
  //   - LOWER_DETAIL - return patch matching pos, but possible with lower detail
  //                    than specified. If pos falls outside the current patch root,
  //                    new nodes are appended to the top of patch tree to match pos.


  // handle requests for higher level
  if (relativeLevel > 0)
    if (parent)  return parent->getPatchHierarchRelative(pos, relativeLevel-1, policy);
    else {
      if (!engine) {
        SoDebugError::post("TePatch::getPatchHierarchRelative",
                           "The patch must be appended to the TePatch tree\n"
                           "before the call of this function.");
        return NULL;
      }
      engine->advanceRootToPos(pos);
      return engine->getRootPatch();
    }

  // not intersecting => go up
  if (!area.intersect(pos)) {
    if (parent)

      // find recursivelly
      return parent->getPatchHierarchRelative(pos, relativeLevel-1, policy);

    else {

      // this item has no parent => it is the root
      if (policy == DONT_CREATE)
        return NULL;
      else {
        if (!engine) {
          SoDebugError::post("TePatch::getPatchHierarchRelative",
                             "The patch must be appended to the TePatch tree\n"
                             "before the call of this function.");
          return NULL;
        }
        engine->advanceRootToPos(pos);

        TePatch *root = engine->getRootPatch();
        if (policy == LOWER_DETAIL)
          return root;
        else
          return root->getPatchHierarchRelative(pos, relativeLevel + (this->getLevel()-root->getLevel()), policy);
      }
    }
  } else {

    if (relativeLevel == 0)  return this;

    Direction d = getDirectionXFromVec(pos-area.getCenter());

    if (detail[d])

      // find recursivelly
      return detail[d]->getPatchHierarchRelative(pos, relativeLevel+1, policy);

    else {

      // requested detail is missing => handle it
      switch (policy) {
      case DONT_CREATE:  return NULL;
      case LOWER_DETAIL: return this;
      case CREATE: {
          this->setDetail(d, new TePatch(engine));
          return detail[d]->getPatchHierarchRelative(pos, relativeLevel+1, policy);
        }
      default: assert(0 && "Wrong policy value."); return NULL;
      }
    }
  }
}


//-----------------------------------------------------------------------------
/**
 * Documentation not available.
 *
 * \todo More detailed documentation.
 */
TePatch* TePatch::getPatch(const SbVec2f &pos, const int level, PatchMissedPolicy policy)
{
  return getPatchRelative(pos, level - this->getLevel(), policy);
}


//-----------------------------------------------------------------------------
/**
 * Documentation not available.
 *
 * \todo More detailed documentation.
 * \todo Make some optimization here using neighbour list.
 */
TePatch* TePatch::getPatchRelative(const SbVec2f &pos, const int relativeLevel, PatchMissedPolicy policy)
{
  // FIXME: make some optimization here using neighbour list
  return getPatchHierarchRelative(pos, relativeLevel, policy);
}


//-----------------------------------------------------------------------------
/**
 * Private attribute accessor.
 *
 * \param which Detail selector.
 * \return Specified detail patch.
 *
 * \sa detail
 */
TePatch* TePatch::getDetail(const Direction which)
{
  assert(which>=0 && which<4 && "Wrong which parameter.");

  return detail[which];
}


//-----------------------------------------------------------------------------
/**
 * Documentation not available.
 *
 * \todo More detailed documentation.
 */
void TePatch::setDetail(const Direction which, TePatch *patch)
{
  assert(which>=0 && which<4 && "Wrong which parameter.");

  if (detail[which] == patch)
    return;

  TePatch *tmp = detail[which];

  detail[which] = patch;

  if (patch) {
    patch->engine = this->engine;
    patch->parent = this;

    // optimization: if previous detail exist, use some of
    // its values instead of computing them
    if (tmp) {

      // area
      patch->area = tmp->area;

      // neighbours
      for (int i=0;i<8;i++)
        patch->neighbour[i] = tmp->neighbour[i];

    } else {

      // area
      SbVec2f min,max;

      if (which==SOUTH_WEST || which==NORTH_WEST) {
        min[0] = this->area.getMin()[0];    max[0] = this->area.getCenter()[0];
      } else {
        min[0] = this->area.getCenter()[0]; max[0] = this->area.getMax()[0];
      }

      if (which==SOUTH_WEST || which==SOUTH_EAST) {
        min[1] = this->area.getMin()[1];    max[1] = this->area.getCenter()[1];
      } else {
        min[1] = this->area.getCenter()[1]; max[1] = this->area.getMax()[1];
      }

      patch->area = SbBox2f(min,max);

      // neighbours
      patch->updateNeighbours();
    }
  }

  // nullify old detail pointers
  if (tmp != NULL) {
    tmp->engine = NULL;
    tmp->parent = NULL;
    for (int i=0;i<8;i++)
      tmp->neighbour[i] = NULL;
  }
}


//-----------------------------------------------------------------------------
/**
 * Updates local neighbour list and references in that neighbours too.
 *
 * \sa neighbour
 */
void TePatch::updateNeighbours()
{
  int i;
  float size_x, size_y;
  area.getSize(size_x, size_y);
  SbVec2f center = area.getCenter();

  // update my neighbour list
  neighbour[SOUTH_WEST] = getPatchHierarchRelative(center + SbVec2f(-size_x,-size_y), 0, DONT_CREATE);
  neighbour[SOUTH_EAST] = getPatchHierarchRelative(center + SbVec2f( size_x,-size_y), 0, DONT_CREATE);
  neighbour[NORTH_WEST] = getPatchHierarchRelative(center + SbVec2f(-size_x, size_y), 0, DONT_CREATE);
  neighbour[NORTH_EAST] = getPatchHierarchRelative(center + SbVec2f( size_x, size_y), 0, DONT_CREATE);

  neighbour[SOUTH] = getPatchHierarchRelative(center + SbVec2f(    0.f,-size_y), 0, DONT_CREATE);
  neighbour[WEST]  = getPatchHierarchRelative(center + SbVec2f(-size_x,    0.f), 0, DONT_CREATE);
  neighbour[EAST]  = getPatchHierarchRelative(center + SbVec2f( size_x,    0.f), 0, DONT_CREATE);
  neighbour[NORTH] = getPatchHierarchRelative(center + SbVec2f(    0.f,+size_y), 0, DONT_CREATE);

  // update references from neighbours
  for (i=0;i<8;i++)
    if (neighbour[i])
      neighbour[i]->neighbour[getOppositeDirection(i)] = this;
}


//-----------------------------------------------------------------------------
/**
 * Documentation not available.
 *
 * \todo More detailed documentation.
 */
void TePatch::removeDetail(TePatch *patch)
{
  assert(patch->parent == this && "Parent/detail integrity error. This should never happen.");

  Direction d = patch->getDirectionWithinParent();
  if (d<0) {
    SoDebugError::postWarning("TePatch::removeDetail",
                              "Attempt to remove detail that is not present in this patch.");
    return;
  }
  this->setDetail(d, NULL);
}


//-----------------------------------------------------------------------------
/**
 * Documentation not available.
 *
 * \todo More detailed documentation.
 */
void TePatch::makeRoot(const SbBox2f &area, TeEngine *engine)
{
  if (this->parent)
    parent->removeDetail(this);

  this->area = area;
  this->engine = engine;
}


//-----------------------------------------------------------------------------
/**
 * Documentation not available.
 *
 * \todo More detailed documentation.
 */
TePatch::Direction TePatch::getDirectionWithinParent() const
{
  if (!parent)
    return Direction(-1);

  for (int i=0; i<4; i++)
    if (detail[i] == this) return Direction(i);

  return Direction(-1);
}


//-----------------------------------------------------------------------------
/**
 * Documentation not available.
 *
 * \todo More detailed documentation.
 */
TePatch::Direction TePatch::getDetailDirection(const SbVec2f &point) const
{
  if (!area.intersect(point))
    return Direction(-1);

  return getDirectionXFromVec(point - area.getCenter());
}


//-----------------------------------------------------------------------------
/**
 * Private attribute accessor.
 *
 * \param i Buildmap selector (TePatch::WhichHMap should be used here).
 * \return Const pointer to the specified build map.
 *
 * \sa WhichHMap, buildMap, setBuildMap(), buildHMapFromBuildMaps()
 */
const TeHeightMap* TePatch::getBuildMap(const int i) const
{
  assert(i>=0 && i<4 && "Wrong which parameter.");
  return buildMap[i];
}


//-----------------------------------------------------------------------------
/**
 * Private attribute accessor.
 *
 * \param i Build map selector (TePatch::WhichHMap should be used here).
 * \return Pointer to the specified build map.
 *
 * \sa WhichHMap, buildMap, setBuildMap(), buildHMapFromBuildMaps()
 */
TeHeightMap* TePatch::getBuildMap(const int i)
{
  assert(i>=0 && i<4 && "Wrong which parameter.");
  return buildMap[i];
}


//-----------------------------------------------------------------------------
/**
 * Sets single build map.
 *
 * This function checks whether the resolution of the heightmap \a hmap
 * correspondes with the one set in the engine. The resolution must be equal
 * to the one returned by TeEngine::getBuildMapResolution() function. If the
 * resolution is incorrect, nothing happens.
 *
 * Info is propagated to all existing neighbours.
 *
 * \param i Buildmap selector (TePatch::WhichHMap should be used here).
 * \param hmap Pointer to the heightmap.
 *
 * \sa setBuildMaps(), WhichHMap, buildMap, getBuildMap(), buildHMapFromBuildMaps()
 */
void TePatch::setBuildMap(const int i, TeHeightMap *hmap)
{
  assert(i>=0 && i<4 && "Wrong which parameter.");
  if (hmap->getResolution() != engine->getBuildMapResolution()) {
    SoDebugError::post("TePatch::setBuildMap",
                       "BuildMap must have resolution equal to the one returned by\n"
                       "TeEngine::getBuildMapResolution(). The resolution is based on the value\n"
                       "passed to TeEngine constructor. For now, this call is ignored.");
    return;
  }

  buildMap[i] = hmap;

  // propagate info to neighbours
  // - uses short evaluation
  // - second check is maybe redundant and probably will be deleted in the
  //   near future
  switch (i) {
    case 0: if (neighbour[WEST] && !neighbour[WEST]->getBuildMap(1)) neighbour[WEST]->setBuildMap(1, hmap);
            if (neighbour[SOUTH_WEST] && !neighbour[SOUTH_WEST]->getBuildMap(3)) neighbour[SOUTH_WEST]->setBuildMap(3, hmap);
            if (neighbour[SOUTH] && !neighbour[SOUTH]->getBuildMap(2)) neighbour[SOUTH]->setBuildMap(2, hmap);
            break;
    case 1: if (neighbour[SOUTH] && !neighbour[SOUTH]->getBuildMap(3)) neighbour[SOUTH]->setBuildMap(3, hmap);
            if (neighbour[SOUTH_EAST] && !neighbour[SOUTH_EAST]->getBuildMap(2)) neighbour[SOUTH_EAST]->setBuildMap(2, hmap);
            if (neighbour[EAST] && !neighbour[EAST]->getBuildMap(0)) neighbour[EAST]->setBuildMap(0, hmap);
            break;
    case 2: if (neighbour[WEST] && !neighbour[WEST]->getBuildMap(3)) neighbour[WEST]->setBuildMap(3, hmap);
            if (neighbour[NORTH_WEST] && !neighbour[NORTH_WEST]->getBuildMap(1)) neighbour[NORTH_WEST]->setBuildMap(1, hmap);
            if (neighbour[NORTH] && !neighbour[NORTH]->getBuildMap(0)) neighbour[NORTH]->setBuildMap(0, hmap);
            break;
    case 3: if (neighbour[NORTH] && !neighbour[NORTH]->getBuildMap(1)) neighbour[NORTH]->setBuildMap(1, hmap);
            if (neighbour[NORTH_EAST] && !neighbour[NORTH_EAST]->getBuildMap(0)) neighbour[NORTH_EAST]->setBuildMap(0, hmap);
            if (neighbour[EAST] && !neighbour[EAST]->getBuildMap(2)) neighbour[EAST]->setBuildMap(2, hmap);
            break;
  }
}


//-----------------------------------------------------------------------------
/**
 * Sets all build maps in a single step.
 *
 * This function simply calls setBuildMap() four times. All restrictions
 * on the heightmaps resolution are therefore applied too.
 *
 * \param sw Pointer to the heightmap, it will become south-west build map.
 * \param se Pointer to the heightmap, it will become south-east build map.
 * \param nw Pointer to the heightmap, it will become north-west build map.
 * \param ne Pointer to the heightmap, it will become north-east build map.
 *
 * \sa buildMap, getBuildMap(), buildHMapFromBuildMaps()
 */
void TePatch::setBuildMaps(TeHeightMap *sw, TeHeightMap *se, TeHeightMap *nw, TeHeightMap *ne)
{
  setBuildMap(HMAP_SOUTH_WEST, sw);
  setBuildMap(HMAP_SOUTH_EAST, se);
  setBuildMap(HMAP_NORTH_WEST, nw);
  setBuildMap(HMAP_NORTH_EAST, ne);
}


//-----------------------------------------------------------------------------
/**
 * Generates heightmap of this patch using fault-formation algorithm.
 *
 * Which height map will be generated is specified by the \a which parameter.
 *
 * More details about how this algorithm works can be found at 
 * TeHeightMap::generate_FaultFormation().
 *
 * \param which Which height map will be generated selector.
 * \param seed Random seed that defines terrain shape.
 * \param num_faults Number of faults to be generated (number of iterations).
 * \param min_delta Minimal fault height.
 * \param max_delta Maximal fault height.
 *
 * \sa WhichHMap, hmap, buildMap
 */
void TePatch::generateHMapFF(WhichHMap which,
                             const unsigned int seed, const int num_faults,
                             const float min_delta, const float max_delta)
{
  assert(which>=0 && which<5 && "Wrong which parameter.");
  TeHeightMap* &p = (which == 4) ? hmap : buildMap[which];

  if (!p)
    p = new TeHeightMap;

  p->setResolution(engine->getHMapResolution());
  p->generate_FaultFormation(seed, num_faults, min_delta, max_delta);
}


//-----------------------------------------------------------------------------
/**
 * Tries to find build map by looking in the neighbours.
 *
 * We suppose that correct info is in all of them, therefore we grab the
 * first one. Info propagation is assured by setBuildMap() method.
 *
 * The function returns void, the effect is visible right inside the
 * buildMap attribute.
 *
 * \param i Buildmap selector (TePatch::WhichHMap should be used here).
 *
 * \sa WhichHMap, buildMap, buildHMapFromBuildMaps()
 */
void TePatch::findBuildMap(const int i)
{
  switch (i) {
    case 0: if (!buildMap[0]) {
              if (neighbour[WEST] && neighbour[WEST]->getBuildMap(1)) this->setBuildMap(0, neighbour[WEST]->getBuildMap(1));
              else if (neighbour[SOUTH_WEST] && neighbour[SOUTH_WEST]->getBuildMap(3)) this->setBuildMap(0, neighbour[SOUTH_WEST]->getBuildMap(3));
              else if (neighbour[SOUTH] && neighbour[SOUTH]->getBuildMap(2)) this->setBuildMap(0, neighbour[SOUTH]->getBuildMap(2));
            }
            break;
    case 1: if (!buildMap[1]) {
              if (neighbour[SOUTH] && neighbour[SOUTH]->getBuildMap(3)) this->setBuildMap(1, neighbour[SOUTH]->getBuildMap(3));
              else if (neighbour[SOUTH_EAST] && neighbour[SOUTH_EAST]->getBuildMap(2)) this->setBuildMap(1, neighbour[SOUTH_EAST]->getBuildMap(2));
              else if (neighbour[EAST] && neighbour[EAST]->getBuildMap(0)) this->setBuildMap(1, neighbour[EAST]->getBuildMap(0));
            }
            break;
    case 2: if (!buildMap[2]) {
              if (neighbour[WEST] && neighbour[WEST]->getBuildMap(3)) this->setBuildMap(2, neighbour[WEST]->getBuildMap(3));
              else if (neighbour[NORTH_WEST] && neighbour[NORTH_WEST]->getBuildMap(1)) this->setBuildMap(2, neighbour[NORTH_WEST]->getBuildMap(1));
              else if (neighbour[NORTH] && neighbour[NORTH]->getBuildMap(0)) this->setBuildMap(2, neighbour[NORTH]->getBuildMap(0));
            }
            break;
    case 3: if (!buildMap[3]) {
              if (neighbour[NORTH] && neighbour[NORTH]->getBuildMap(1)) this->setBuildMap(3, neighbour[NORTH]->getBuildMap(1));
              else if (neighbour[NORTH_EAST] && neighbour[NORTH_EAST]->getBuildMap(0)) this->setBuildMap(3, neighbour[NORTH_EAST]->getBuildMap(0));
              else if (neighbour[EAST] && neighbour[EAST]->getBuildMap(2)) this->setBuildMap(3, neighbour[EAST]->getBuildMap(2));
            }
            break;
  }
}


//-----------------------------------------------------------------------------
/**
 * High-level function that prepares all build maps.
 *
 * Tries to find build maps in the neighbours. If map is not found, asks
 * engine to generate new one.
 *
 * \sa buildMap, TeEngine::generateHMap(), buildHMapFromBuildMaps()
 */
void TePatch::prepareBuildMaps()
{
  for (int i=0; i<4; i++) {
    this->findBuildMap(i);
    if (!buildMap[i]) {
      engine->faultFormation.seed = (unsigned int)time(NULL)+(rand()%10000);
      this->setBuildMap(i, engine->generateHMap(engine->getBuildMapResolution()));
    }
  }
}


//-----------------------------------------------------------------------------
/**
 * Computes patch height map from four build maps.
 *
 * All build maps have to be set before calling this method.
 *
 * \sa buildMap, setBuildMap()
 *
 * \todo More detailed documentation.
 */
void TePatch::buildHMapFromBuildMaps()
{
  if (!buildMap[0] || !buildMap[1] || !buildMap[2] || !buildMap[3]) {
    SoDebugError::post("TePatch::buildHMapFromBuildMaps",
                       "Not all build maps were set. Can not build HeightMap.");
    return;
  }

  SbVec2s h_res = engine->getHMapResolution();      // height map resolution
  SbVec2s b_res = engine->getBuildMapResolution();  // build map resolution

  assert((buildMap[0]->getResolution() == b_res) && (buildMap[1]->getResolution() == b_res) &&
         (buildMap[2]->getResolution() == b_res) && (buildMap[3]->getResolution() == b_res) &&
         "buildMaps resolutions are not the same.");

  int shift = b_res[0] - h_res[0];

  if (hmap)
    delete hmap;

  hmap = new TeHeightMap(h_res);

  float *dest = hmap->startEditing();
  const float *m1 = &buildMap[0]->getValues()[b_res[0]*(h_res[1]-1) + h_res[0]-1];
  const float *m2 = &buildMap[1]->getValues()[b_res[0]*(h_res[1]-1)];
  const float *m3 = &buildMap[2]->getValues()[h_res[0]-1];
  const float *m4 = &buildMap[3]->getValues()[0];

  int x,y;
  int index;
  float qf; // koeficient "going forward" - increasing from 0 on the first row to 1 on the last row
  float qb; // koeficient "going backward" - decreasing from 1 on the first row to 0 on the last row
  float q1; // koeficient decreasing from qb to 0 on each row
  float q2; // koeficient increasing from 0 to qb on each row
  float q3; // koeficient decreasing from qf to 0 on each row
  float q4; // koeficient increasing from 0 to qf on each row

  float dy = 1.f/float(h_res[1]-1);
  float dxf,dxb; // size of q1,q2,q3, and q4 step

  qf = 0.f;
  qb = 1.f;
  index = 0;

  for (y=0; y<h_res[1]; y++) {

    q1 = qb;
    q2 = 0.f;
    q3 = qf;
    q4 = 0.f;
    dxf = qf / (h_res[0]-1);
    dxb = qb / (h_res[0]-1);

    for (x=0; x<h_res[0]; x++) {

      *dest++ = (*m1++)*q1 + (*m2++)*q2 + (*m3++)*q3 + (*m4++)*q4;

      q1 -= dxb;
      q2 += dxb;
      q3 -= dxf;
      q4 += dxf;
    }

    m1 += shift;
    m2 += shift;
    m3 += shift;
    m4 += shift;

    qf += dy;
    qb -= dy;
  }

  hmap->finishEditing();
}


//-----------------------------------------------------------------------------
/**
 * Validates seamMap pointer in desired direction if possible.
 *
 * Tries to look for the desired map in the neighbours. If not found, tries
 * to compute it and propagate the info to all interested neighbours. This
 * assures that each seam map will be computed only once. User does not
 * need to worry about the way how the data are obtained. This function
 * says if the private seam map pointer is valid or not.
 *
 * Note that properly set hmap attributes of all necessary neighbours are
 * needed to succesfully complete this task.
 *
 * \param which Seam map selector.
 * \return TRUE if the seam map pointer is valid and the graph can be created.
 *
 * \sa seamMap, getSeamMap()
 */
SbBool TePatch::generateSeamMap(const Direction which)
{
  // map already set
  if (seamMap[which]) return TRUE;

  // do we have all neighbours and hmaps ready?
  if (!this->hmap) return FALSE;
  if (which == SOUTH || which == WEST || which == EAST || which == NORTH) {
    if (!neighbour[which]) return FALSE;
    else if (!neighbour[which]->hmap) return FALSE;
  }
  else {
    if (!neighbour[which]) return FALSE;
    else if (!neighbour[which]->hmap) return FALSE;
    switch (which) {
      case SOUTH_WEST: if (!neighbour[SOUTH] || !neighbour[WEST]) return FALSE;
                       else if (!neighbour[SOUTH]->hmap || !neighbour[WEST]->hmap) return FALSE;
                       break;
      case SOUTH_EAST: if (!neighbour[SOUTH] || !neighbour[EAST]) return FALSE;
                       else if (!neighbour[SOUTH]->hmap || !neighbour[EAST]->hmap) return FALSE;
                       break;
      case NORTH_WEST: if (!neighbour[NORTH] || !neighbour[WEST]) return FALSE;
                       else if (!neighbour[NORTH]->hmap || !neighbour[WEST]->hmap) return FALSE;
                       break;
      case NORTH_EAST: if (!neighbour[NORTH] || !neighbour[EAST]) return FALSE;
                       else if (!neighbour[NORTH]->hmap || !neighbour[EAST]->hmap) return FALSE;
                       break;
    }
  }

  // if yes, try to look for the map in one of the neighbours (we rely on
  // correct info to be in all of them)
  TeHeightMap *tmp = neighbour[which]->seamMap[getOppositeDirection(which)];
  if (tmp) {
    seamMap[which] = tmp;
    return TRUE;
  }

  int x, y = 0;
  int resx = engine->getHMapResolution()[0];
  int resy = engine->getHMapResolution()[1];

  // if not found, compute map from neighbour data and update neighbour's
  // seamMap-pointer
  switch (which) {
    case SOUTH_WEST: tmp = new TeHeightMap(5, 5);

                     for (y=2; y<5; y++)
                       for (x=0; x<3; x++)
                         tmp->setValue(x, y, neighbour[WEST]->hmap->getValue(resx-3+x, y-2));

                     for (y=2; y<5; y++)
                       for (x=3; x<5; x++)
                         tmp->setValue(x, y, this->hmap->getValue(x-2, y-2));

                     for (y=0; y<2; y++)
                       for (x=0; x<3; x++)
                         tmp->setValue(x, y, neighbour[SOUTH_WEST]->hmap->getValue(resx-3+x, resy-3+y));

                     for (y=0; y<2; y++)
                       for (x=3; x<5; x++)
                         tmp->setValue(x, y, neighbour[SOUTH]->hmap->getValue(x-2, resy-3+y));

                     neighbour[WEST]->setSeamMap(SOUTH_EAST, tmp);
                     neighbour[SOUTH]->setSeamMap(NORTH_WEST, tmp);
                     break;

    case SOUTH_EAST: tmp = new TeHeightMap(5, 5);

                     for (y=2; y<5; y++)
                       for (x=0; x<3; x++)
                         tmp->setValue(x, y, this->hmap->getValue(resx-3+x, y-2));

                     for (y=2; y<5; y++)
                       for (x=3; x<5; x++)
                         tmp->setValue(x, y, neighbour[EAST]->hmap->getValue(x-2, y-2));

                     for (y=0; y<2; y++)
                       for (x=0; x<3; x++)
                         tmp->setValue(x, y, neighbour[SOUTH]->hmap->getValue(resx-3+x, resy-3+y));

                     for (y=0; y<2; y++)
                       for (x=3; x<5; x++)
                         tmp->setValue(x, y, neighbour[SOUTH_EAST]->hmap->getValue(x-2, resy-3+y));

                     neighbour[EAST]->setSeamMap(SOUTH_WEST, tmp);
                     neighbour[SOUTH]->setSeamMap(NORTH_EAST, tmp);
                     break;

    case NORTH_WEST: tmp = new TeHeightMap(5, 5);

                     for (y=2; y<5; y++)
                       for (x=0; x<3; x++)
                         tmp->setValue(x, y, neighbour[NORTH_WEST]->hmap->getValue(resx-3+x, y-2));

                     for (y=2; y<5; y++)
                       for (x=3; x<5; x++)
                         tmp->setValue(x, y, neighbour[NORTH]->hmap->getValue(x-2, y-2));

                     for (y=0; y<2; y++)
                       for (x=0; x<3; x++)
                         tmp->setValue(x, y, neighbour[WEST]->hmap->getValue(resx-3+x, resy-3+y));

                     for (y=0; y<2; y++)
                       for (x=3; x<5; x++)
                         tmp->setValue(x, y, this->hmap->getValue(x-2, resy-3+y));

                     neighbour[NORTH]->setSeamMap(SOUTH_WEST, tmp);
                     neighbour[WEST]->setSeamMap(NORTH_EAST, tmp);
                     break;

    case NORTH_EAST: tmp = new TeHeightMap(5, 5);

                     for (y=2; y<5; y++)
                       for (x=0; x<3; x++)
                         tmp->setValue(x, y, neighbour[NORTH]->hmap->getValue(resx-3+x, y-2));

                     for (y=2; y<5; y++)
                       for (x=3; x<5; x++)
                         tmp->setValue(x, y, neighbour[NORTH_EAST]->hmap->getValue(x-2, y-2));

                     for (y=0; y<2; y++)
                       for (x=0; x<3; x++)
                         tmp->setValue(x, y, this->hmap->getValue(resx-3+x, resy-3+y));

                     for (y=0; y<2; y++)
                       for (x=3; x<5; x++)
                         tmp->setValue(x, y, neighbour[EAST]->hmap->getValue(x-2, resy-3+y));

                     neighbour[NORTH]->setSeamMap(SOUTH_EAST, tmp);
                     neighbour[EAST]->setSeamMap(NORTH_WEST, tmp);
                     break;

         case SOUTH: tmp = new TeHeightMap(SbVec2s(engine->getHMapResolution()[0], 5));
                     tmp->copyFrom(neighbour[which]->hmap,
                                   SbVec2s(0, engine->getHMapResolution()[1]-3),
                                   SbVec2s(engine->getHMapResolution()[0], 3));
                     tmp->copyFrom(this->hmap,
                                   SbVec2s(0, 1),
                                   SbVec2s(engine->getHMapResolution()[0], 2),
                                   SbVec2s(0, 3));
                     break;

          case WEST: tmp = new TeHeightMap(SbVec2s(5, engine->getHMapResolution()[1]));
                     tmp->copyFrom(neighbour[which]->hmap,
                                   SbVec2s(engine->getHMapResolution()[0]-3, 0),
                                   SbVec2s(3, engine->getHMapResolution()[1]));
                     tmp->copyFrom(this->hmap,
                                   SbVec2s(1, 0),
                                   SbVec2s(2, engine->getHMapResolution()[1]),
                                   SbVec2s(3, 0));
                     break;

          case EAST: tmp = new TeHeightMap(SbVec2s(5, engine->getHMapResolution()[1]));
                     tmp->copyFrom(this->hmap,
                                   SbVec2s(engine->getHMapResolution()[0]-3, 0),
                                   SbVec2s(3, engine->getHMapResolution()[1]));
                     tmp->copyFrom(neighbour[which]->hmap,
                                   SbVec2s(1, 0),
                                   SbVec2s(2, engine->getHMapResolution()[1]),
                                   SbVec2s(3, 0));
                     break;

         case NORTH: tmp = new TeHeightMap(SbVec2s(engine->getHMapResolution()[0], 5));
                     tmp->copyFrom(this->hmap,
                                   SbVec2s(0, engine->getHMapResolution()[1]-3),
                                   SbVec2s(engine->getHMapResolution()[0], 3));
                     tmp->copyFrom( neighbour[which]->hmap,
                                   SbVec2s(0, 1),
                                   SbVec2s(engine->getHMapResolution()[0], 2),
                                   SbVec2s(0, 3));
                     break;
  }

  // always
  this->setSeamMap(which, tmp);
  neighbour[which]->setSeamMap(Direction(getOppositeDirection(which)), tmp);
  // more inside the switch above

  return TRUE;
}


//-----------------------------------------------------------------------------
/**
 * Calculates normal at the specified point in the heightmap.
 *
 * \param map Pointer to the height map.
 * \param x X-coordinate of the point.
 * \param y Y-coordinate of the point.
 * \param sx ???
 * \param sy ???
 * \return Normal at the specified point (\a x, \a y).
 *
 * \todo More detailed documentation. Parameters sx, sy ???
 */
SbVec3f TePatch::calculateNormal(const TeHeightMap *map, const int x,
                                 const int y, const float sx, const float sy)
{
  float dx,dy;
  int resx,resy;
  map->getResolution(resx,resy);
  const float *values = map->getValues();

  // hel v ose X
  if (x>=1 && x<=resx-2)
    // Vpoet ze dvou sousednch bod
    dx = (values[y*resx + x-1] - values[y*resx + x+1]) / 2.f;
  else
    // Vpoet na okraji mapy
    if (x<1)  dx = values[y*resx + x] - values[y*resx + x+1];
    else  dx = values[y*resx + x-1] - values[y*resx + x];

  // hel v ose Y
  if (y>=1 && y<=resy-2)
    dy = (values[(y-1)*resx + x] - values[(y+1)*resx + x]) / 2.f;
  else
    if (y<1)  dy = values[y*resx + x] - values[(y+1)*resx + x];
    else  dy = values[(y-1)*resx + x] - values[y*resx + x];

  SbVec3f r(dx/sx, dy/sy, 1.f);
  r.normalize();

  return r;
};


#define TPATCH_SEA_LEVEL (float)180
#define TPATCH_ROCK_LEVEL (float)195
#define TPATCH_SNOW_LEVEL (float)200


//-----------------------------------------------------------------------------
/**
 * This function is used to select "colour" of the terrain of given height.
 * This function is for internal use only and uses #defines that can be found
 * right above it in the source code.
 *
 * \param value Number representing height of the point.
 * \return Index to the material node.
 *
 * \todo This function should NOT use #defined values. It should guess the "levels" somehow.
 */
unsigned short TePatch::selectMaterial(float value)
{
  if ( value < TPATCH_ROCK_LEVEL ) return 1;
  else if ( value < TPATCH_SNOW_LEVEL ) return 2;
  else return 3;
}



//-----------------------------------------------------------------------------
/**
 * The function creates scene graph of the specified area of height map given
 * by the \a map parameter. The parameters \a min and \a max specifies the
 * rectangle of the height map that will be used for scene graph generation.
 * The parameter \a area gives the world-space area that will be occupied
 * by the resulting scene graph.
 *
 * Please note that normal generation algorithm tries to access height map
 * values one point over the rectangle specified by \a min and \a max
 * parameters. If these values can not be accessed, the normals are computed
 * without them, but this usually results in artefacts like visible seams
 * between patches.
 *
 * \param map Pointer to the source heightmap.
 * \param min Min corner of selected heightmap region.
 * \param max Max corner of selected heightmap region.
 * \param area World-space area that will be occupied by the scene graph.
 * \return Scene graph that represents the terrain defined by selected
 *         heightmap values.
 */
SoSeparator* TePatch::generateGraphFromSubMap(const TeHeightMap *map,
                                              const SbVec2s &min,
                                              const SbVec2s &max,
                                              const SbBox2f &area)
{
  assert(map && "Passed NULL height map.");
  assert(min[0] >= 0 && min[1] >= 0 && max[0] <= map->getResolution()[0] &&
         max[1] <= map->getResolution()[1] && "Wrong min or max argument.");

  SoSeparator *root = new SoSeparator;

  SoTranslation *trans = new SoTranslation;
  trans->translation.setValue(area.getMin()[0], area.getMin()[1], 0.f);
  root->addChild(trans);

  // Teren
  // definice materialu (barvy) krajiny
  SoMaterial *material = new SoMaterial;
  // voda
  material->ambientColor.set1Value(0, SbColor(0.9f, 0.9f, 0.9f));
  material->diffuseColor.set1Value(0, SbColor(0.0f, 0.0f, 0.6f));
  // trava
  material->ambientColor.set1Value(1, SbColor(0.9f, 0.9f, 0.9f));
  material->diffuseColor.set1Value(1, SbColor(0.0f, 0.6f, 0.0f));
  // skaly
  material->ambientColor.set1Value(2, SbColor(0.9f, 0.9f, 0.9f));
  material->diffuseColor.set1Value(2, SbColor(0.6f, 0.3f, 0.0f));
  // snih
  material->ambientColor.set1Value(3, SbColor(0.9f, 0.9f, 0.9f));
  material->diffuseColor.set1Value(3, SbColor(0.6f, 0.6f, 0.6f));
  root->addChild(material);

  // binding pro normaly - v nasem pripade PER_VERTEX_INDEXED
  SoNormalBinding *nbind = new SoNormalBinding;
  nbind->value.setValue(SoNormalBinding::PER_VERTEX_INDEXED);
  root->addChild(nbind);

  // binding pro materialy - v nasem pripade PER_VERTEX_INDEXED
  SoMaterialBinding *mbind = new SoMaterialBinding;
  mbind->value.setValue(SoMaterialBinding::PER_VERTEX_INDEXED);
  root->addChild(mbind);


  // vytvor nody pro souradnice a normaly
  SoCoordinate3 *coords = new SoCoordinate3;
  SoNormal *normals = new SoNormal;

  // nastavime pocet souradnic a normal, ktere budeme generovat
  int mapResx = map->getResolution()[0];
  int resx = max[0] - min[0];
  int resy = max[1] - min[1];
  int numVertices = resx*resy;
  coords->point.setNum( numVertices );
  normals->vector.setNum( numVertices );

  // ukazatele na souradnice a normaly
  SbVec3f *c = coords->point.startEditing();
  SbVec3f *n = normals->vector.startEditing();

  int x, y;
  SbVec2f areaSize = area.getMax() - area.getMin();
  float sx = areaSize[0] / (resx-1);
  float sy = areaSize[1] / (resy-1);
  const float *values = map->getValues();

  // Kazdy ctverec povrchu je tvoren dvema trojuhelniky,
  // vertexy odpovidajici jednotlivym pixelum na mape
  int minx = min[0];
  int miny = min[1];
  for (y = miny; y < max[1]; y++) {
    for (x = minx; x < max[0]; x++) {
      *(c++) = SbVec3f(float(x-minx)*sx, float(y-miny)*sy, values[y*mapResx+x]);
      *(n++) = calculateNormal(map, x,y, sx,sy);
    }
  }

  // zaradime vytvorene uzly do grafu sceny
  coords->point.finishEditing();
  normals->vector.finishEditing();
  root->addChild(normals);
  root->addChild(coords);

  // konfigurace SoIndexedTriangleStripSetu
  SoIndexedTriangleStripSet *stripSet = new SoIndexedTriangleStripSet;

  // nastavime pocet indexu, ktere budeme generovat
  int numIndices = 5*(resx-1)*(resy-1);
  stripSet->coordIndex.setNum( numIndices );
  stripSet->materialIndex.setNum( numIndices );

  // ukazatele na indexy (normalIndexy jsou shodne s coordIndexy,
  // nemusi se nastavovat)
  int32_t *ci = stripSet->coordIndex.startEditing();
  int32_t *mi = stripSet->materialIndex.startEditing();

  int32_t index = 0;
  long distance = 0;

  // kazda dlazdice terenu je tvorena dvema trojuhelniky
  // rendrovanymi jako jeden triangle strip,
  // zlom mezi trojuhelniky je generovan podle smeru vetsi zmeny vysky v dlazdici
  index = 0;
  int i1 = 0;
  int i2 = resx;
  for (int strip = 0; strip < resy-1; strip++) {
    for (int column = 0; column < resx-1; column++) {
      // potej zmnu v jednotlivch smrech dladice,
      // abychom zjistili, ve kterm smru bude dladice rozdlena.
      bool orderChange = fabsf(values[i2+1]-values[i1]) < fabsf(values[i2]-values[i1+1]);
      if (orderChange) {
        int tmp = i2;
        i2 = i1;
        i1 = tmp;
      }

      ci[index] = i1;
      mi[index] = selectMaterial(coords->point[i1][2]);
      index++;
      ci[index] = i2;
      mi[index] = selectMaterial(coords->point[i2][2]);
      index++;
      i1++;
      i2++;
      ci[index] = i1;
      mi[index] = selectMaterial(coords->point[i1][2]);
      index++;
      ci[index] = i2;
      mi[index] = selectMaterial(coords->point[i2][2]);
      index++;
      ci[index] = -1; // konec Triangle_stripu
      mi[index] = -1;
      index++;
    }
    i1++;
    i2++;
  }

  // zaradime vytvoreny StripSet do grafu sceny
  stripSet->coordIndex.finishEditing();
  stripSet->materialIndex.finishEditing();
  root->addChild(stripSet);

  //---------------------------------------------------------------------------
  // Vodni hladina

  // vytvor nody pro souradnice a normaly
  SoCoordinate3 *waterCoords = new SoCoordinate3;
  SoNormal *waterNormals = new SoNormal;

  // nastavime pocet souradnic a normal
  waterCoords->point.setNum(4);
  waterNormals->vector.setNum(4);

  // ukazatele na souradnice a normaly
  c = waterCoords->point.startEditing();
  n = waterNormals->vector.startEditing();

  float width, height = 0.0f;
  area.getSize(width, height);

  *(c++) = SbVec3f( 0.0f, 0.0f, TPATCH_SEA_LEVEL );
  *(c++) = SbVec3f( (float)width, 0.0f, TPATCH_SEA_LEVEL );
  *(c++) = SbVec3f( 0.0f, (float)height, TPATCH_SEA_LEVEL );
  *(c++) = SbVec3f( (float)width, (float)height, TPATCH_SEA_LEVEL );
  *(n++) = SbVec3f(0.0f, 0.0f, 1.0f);
  *(n++) = SbVec3f(0.0f, 0.0f, 1.0f);
  *(n++) = SbVec3f(0.0f, 0.0f, 1.0f);
  *(n++) = SbVec3f(0.0f, 0.0f, 1.0f);

  // zaradime vytvorene uzly do grafu sceny
  waterCoords->point.finishEditing();
  waterNormals->vector.finishEditing();
  root->addChild(waterNormals);
  root->addChild(waterCoords);

  // konfigurace SoIndexedTriangleStripSetu
  SoIndexedTriangleStripSet *waterStripSet = new SoIndexedTriangleStripSet;
  
  // nastavime pocet indexu, ktere budeme generovat
  waterStripSet->coordIndex.setNum( 4 );
  waterStripSet->normalIndex.setNum( 4 );
  waterStripSet->materialIndex.setNum( 4 );

  // ukazatele na indexy (normalIndexy jsou shodne s coordIndexy,
  // nemusi se nastavovat)
  ci = waterStripSet->coordIndex.startEditing();
  mi = waterStripSet->materialIndex.startEditing();

  ci[0] = 0;
  ci[1] = 1;
  ci[2] = 2;
  ci[3] = 3;

  mi[0] = 0;
  mi[1] = 0;
  mi[2] = 0;
  mi[3] = 0;

  // zaradime vytvoreny StripSet do grafu sceny
  waterStripSet->coordIndex.finishEditing();
  waterStripSet->materialIndex.finishEditing();
  root->addChild(waterStripSet);

  //---------------------------------------------------------------------------

  root->ref();
  return root;
}


//-----------------------------------------------------------------------------
/**
 * Generates "raw" scene graph of the patch.
 *
 * All heightmap values will be used and therefore there won't be gap between
 * patches but visible seam artefacts will be probably over the landscape.
 *
 * Note, that hmap has to be set before calling this function. If it isn't set,
 * empty scene graph is generated.
 *
 * It is recommended to use generatePatchGraph() instead and later fill gaps
 * between paches with seams.
 *
 * \return "Raw" scene graph of the patch. Empty scene, if hmap is not set.
 *
 * \sa getGraph(), generatePatchGraph(), getPatchGraph()
 */
SoSeparator* TePatch::generateGraph()
{
  if (!hmap) {
    SoDebugError::post("TePatch::generateGraph",
                       "The patch does not have set its height map. Empty scene graph is generated.");

    return new SoSeparator;
  }

  return generateGraphFromSubMap(hmap, SbVec2s(0,0), hmap->getResolution(), area);
}


//-----------------------------------------------------------------------------
/**
 * Generates scene graph of the patch.
 *
 * Not all heightmap values will be used (only those, for that the normal
 * can be computed) and therefore there will be gap between patches that has
 * to be filled with seam.
 *
 * Note, that hmap has to be set before calling this function. If it isn't set,
 * empty scene graph is generated.
 *
 * \return Scene graph of the patch. Empty scene, if hmap is not set.
 *
 * \sa getPatchGraph()
 */
SoSeparator* TePatch::generatePatchGraph()
{
  if (!hmap) {
    SoDebugError::post("TePatch::generatePatchGraph",
                       "The patch does not have set its height map. Empty scene graph is generated.");

    return new SoSeparator;
  }

  float sx,sy;
  area.getSize(sx,sy);
  SbVec2f step(sx/hmap->getResolution()[0], sy/hmap->getResolution()[1]);
  return generateGraphFromSubMap(hmap, SbVec2s(1,1), hmap->getResolution() - SbVec2s(1,1),
                                 SbBox2f(area.getMin()+step, area.getMax()-step));
}


//-----------------------------------------------------------------------------
/**
 * Generates scene graph of the selected seam.
 *
 * If the seam heightmap is not set it tries to generate it, so it is not
 * necessary to call generateSeamMap() before calling this function.
 *
 * \param which Seam selector.
 * \return Scene graph of the seam. NULL, if it was impossible to compute
 *         seam heightmap.
 *
 * \sa getSeamGraph()
 */
SoSeparator* TePatch::generateSeamGraph(const Direction which)
{
  assert(which>=0 && which<8 && "Wrong which parameter.");

  if (!this->generateSeamMap(which)) return NULL;

  float sx,sy;
  area.getSize(sx,sy);
  float stepx = sx/hmap->getResolution()[0];
  float stepy = sy/hmap->getResolution()[1];

  float aminx = area.getMin()[0];
  float aminy = area.getMin()[1];
  float amaxx = area.getMax()[0];
  float amaxy = area.getMax()[1];

  float x1 = aminx - stepx;
  float x2 = aminx + stepx;
  float x3 = amaxx - stepx;
  float x4 = amaxx + stepx;
  float y1 = aminy - stepy;
  float y2 = aminy + stepy;
  float y3 = amaxy - stepy;
  float y4 = amaxy + stepy;

  SbVec2f min,max;

  switch (which) {
    case SOUTH_WEST: min.setValue(x1,y1); max.setValue(x2,y2); break;
    case SOUTH_EAST: min.setValue(x3,y1); max.setValue(x4,y2); break;
    case NORTH_WEST: min.setValue(x1,y3); max.setValue(x2,y4); break;
    case NORTH_EAST: min.setValue(x3,y3); max.setValue(x4,y4); break;
         case SOUTH: min.setValue(x2,y1); max.setValue(x3,y2); break;
          case WEST: min.setValue(x1,y2); max.setValue(x2,y3); break;
          case EAST: min.setValue(x3,y2); max.setValue(x4,y3); break;
         case NORTH: min.setValue(x2,y3); max.setValue(x3,y4); break;
  }

  return generateGraphFromSubMap(seamMap[which], SbVec2s(1,1),
                                 seamMap[which]->getResolution() - SbVec2s(1,1),
                                 SbBox2f(min,max));
}


//-----------------------------------------------------------------------------
/**
 * High-level "raw" graph request.
 *
 * All heightmap values will be used and therefore there won't be gap between
 * patches but visible seam artefacts will be probably over the landscape.
 * This function does NOT use caching in the TePatch::graph attribute.
 *
 * This rendering mode is obsolete and is recommended to use getPatchGraph()
 * and getSeamGraph() methods instead.
 *
 * \return "Raw" patch graph, or empty scene if height map is not set.
 *
 * \sa getPatchGraph()
 */
SoSeparator *TePatch::getGraph()
{
  return this->generateGraph();
}


//-----------------------------------------------------------------------------
/**
 * High-level graph request.
 *
 * Only those heightmap values, for that we can compute normals will be used,
 * and therefore a gap between patches will occure that has to be filled
 * with seam. The seams may be obtained using getSeamGraph() method.
 * Produced landscape won't have any artefacts along the seam borders then.
 *
 * This function uses caching in the TePatch::graph attribute, so graph is
 * generated only once and is stored for the next time it is needed. Then
 * it will be available very quickly.
 *
 * \return Patch graph, or empty scene if height map is not set.
 *
 * \sa getSeamGraph()
 */
SoSeparator *TePatch::getPatchGraph()
{
  if (!graph) graph = generatePatchGraph();
  return graph;
}


//-----------------------------------------------------------------------------
/**
 * High-level seam graph request.
 *
 * Seams are used to fill gaps between patches that occure because of
 * the normal problem when using getPatchGraph() method.
 *
 * This function uses caching in the TePatch::seamGraph attribute, so graph is
 * generated only once and is stored for the next time it is needed. Then
 * it will be available very quickly.
 *
 * Information about the seam is propagated to all affected neighbours
 * allowing them to share the same data and never do the same work twice.
 *
 * \return Seam graph, or empty scene if seam's height map is not set.
 *
 * \sa getPatchGraph()
 */
SoSeparator *TePatch::getSeamGraph(const Direction which)
{
  SoSeparator *tmp;

  // not cached, try to generate
  if (!seamGraph[which]) tmp = generateSeamGraph(which);

  // if OK, propagate info to local and neighbours' cache
  if (tmp) {
    // always this...
    seamGraph[which] = tmp;
    neighbour[which]->seamGraph[getOppositeDirection(which)] = tmp;
    // more neighbours are affected when generating in these directions:
    switch (which) {
      case SOUTH_WEST: neighbour[WEST]->seamGraph[SOUTH_EAST] = tmp;
                       neighbour[SOUTH]->seamGraph[NORTH_WEST] = tmp;
                       break;
      case SOUTH_EAST: neighbour[EAST]->seamGraph[SOUTH_WEST] = tmp;
                       neighbour[SOUTH]->seamGraph[NORTH_EAST] = tmp;
                       break;
      case NORTH_WEST: neighbour[NORTH]->seamGraph[SOUTH_WEST] = tmp;
                       neighbour[WEST]->seamGraph[NORTH_EAST] = tmp;
                       break;
      case NORTH_EAST: neighbour[NORTH]->seamGraph[SOUTH_EAST] = tmp;
                       neighbour[EAST]->seamGraph[NORTH_WEST] = tmp;
              default: break;
    } // switch
    return seamGraph[which];
  }
  else {
    SoDebugError::post("TePatch::getSeamGraph",
                       "Specified seam map is not available. Empty scene graph will be returned.");
    return new SoSeparator;
  }
}
