//##################################################
//# $Id: TeEngine.cpp 19 2005-03-01 15:15:06Z peciva $

/*****************************************************************************\
 *
 * TeEngine.cpp
 *
 * TeEngine - terrain engine class
 *
 * Authors: Martin Havlek (xhavli15 AT stud.fit.vutbr.cz)
 *          PCJohn (peciva AT fit.vutbr.cz)
 * Contributors: 
 *
 * ----------------------------------------------------------------------------
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, authors will kindly welcome
 * if you give them credit and keep their names with their source code.
 *
\*****************************************************************************/


#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/errors/SoDebugError.h>

#include <cstdlib>    // NULL
#include <ctime>      // time

#include "TeEngine.h"


//----------------------------------------------------------------------------
/**
 * Default constructor.
 *
 * Does some default choices, that can be modified later.
 * TeEngine::FAULT_FORMATION is selected as a generator and TeEngine::LINEAR
 * as a filter.
 *
 * \param hmapResolution Resolution of the patches' heightmaps. If left blank,
 *        default value is used.
 * \param patch0Size Area occupied by the patch in the world-space. If left
 *        blank, default value is used.
 */
TeEngine::TeEngine(SbVec2s hmapResolution, SbVec2f patch0Size) : root(NULL),
  currGen(FAULT_FORMATION), currFilter(LINEAR), HUDCamera(NULL), HUDRoot(NULL),
  sceneCamera(NULL), sceneRoot(NULL), activeObject(NULL), fps(0.0f),
  tasksInQueue(0), patchesNum(0)
{
  this->hmapResolution = hmapResolution;
  this->patch0Size = patch0Size;
  generator = &faultFormation;
}


//----------------------------------------------------------------------------
/**
 * Destructor.
 */
TeEngine::~TeEngine()
{
  delete root;
}


//-----------------------------------------------------------------------------
/**
 * Private attribute accessor.
 *
 * \return Patch height map resolution.
 *
 * \sa hmapResolution
 */
const SbVec2s& TeEngine::getHMapResolution() const
{
  return hmapResolution;
}


//-----------------------------------------------------------------------------
/**
 * Computes patch build maps resolution.
 *
 * Build maps have to have double sized dimensions than the resulting height
 * map (minus 1).
 *
 * \return Patch build map resolution.
 */
SbVec2s TeEngine::getBuildMapResolution() const
{
  return hmapResolution * 2 - SbVec2s(1,1);
}


//-----------------------------------------------------------------------------
/**
 * Private attribute accessor.
 *
 * \return Patch height map resolution.
 *
 * \sa patch0Size
 */
const SbVec2f& TeEngine::getPatch0Size() const
{
  return patch0Size;
}


//-----------------------------------------------------------------------------
/**
 * Private attribute accessor.
 *
 * \return Currently selected generation method.
 *
 * \sa currGen, setCurrGen()
 */
TeEngine::GenMethod TeEngine::getCurrGen()
{
  return currGen;
}


//-----------------------------------------------------------------------------
/**
 * Sets currently selected generation method to \a m
 *
 * This method will be used when new task is scheduled or when generateHMap()
 * method is called.
 *
 * \param m Selected generation method.
 *
 * \sa currGen, getCurrGen(), schedule()
 */
void TeEngine::setCurrGen(GenMethod m)
{
  currGen = m;
}


//-----------------------------------------------------------------------------
/**
 * Private attribute accessor.
 *
 * \return Currently selected filtration method.
 *
 * \sa currFilter, setCurrFilter()
 */
TeEngine::FilterMethod TeEngine::getCurrFilter()
{
  return currFilter;
}


//-----------------------------------------------------------------------------
/**
 * Sets currently selected filtration method to \a m
 *
 * This method will be used when new task is scheduled or when generateHMap()
 * method is called.
 *
 * \param m Selected filtration method.
 *
 * \sa currFilter, getCurrFilter(), schedule()
 */
void TeEngine::setCurrFilter(FilterMethod m)
{
  currFilter = m;
}


//-----------------------------------------------------------------------------
/**
 * Adds object into the engine.
 *
 * \param obj Object to be added to the object list.
 *
 * \sa objectList, getObject(), removeObject()
 */
void TeEngine::addObject(TeObject *obj)
{
  obj->engine = this;
  objectList.append(obj);
  sceneRoot->addChild(obj->getSceneGraph());
}


//-----------------------------------------------------------------------------
/**
 * Gets object from the object list.
 *
 * \param idx Index into the object list.
 * \return Object at index \a idx in the object list
 *
 * \sa objectList, addObject(), removeObject()
 */
TeObject *TeEngine::getObject(int idx)
{
  return objectList[idx];
}


//-----------------------------------------------------------------------------
/**
 * Removes the \a object from engine.
 *
 * \param obj Object to be removed from the object list.
 *
 * \sa objectList, addObject(), getObject()
 */
void TeEngine::removeObject(TeObject *obj)
{
  objectList.removeItem(obj);
}


//-----------------------------------------------------------------------------
/**
 * Sets active object.
 *
 * \param obj Object to be set as active.
 *
 * \sa activeObject, getActiveObject()
 */
void TeEngine::setActiveObject(TeObject *obj)
{
  activeObject = obj;
}


//-----------------------------------------------------------------------------
/**
 * Gets active object.
 *
 * \return Currently active object.
 *
 * \sa activeObject, setActiveObject()
 */
TeObject *TeEngine::getActiveObject()
{
  return activeObject;
}


//-----------------------------------------------------------------------------
/**
 * Sets whole scene root.
 *
 * All objects managed by the engine will be added to the scene as children
 * of this node.
 *
 * \param sep Selected scene root.
 *
 * \sa sceneRoot, getSceneRoot()
 */
void TeEngine::setSceneRoot(SoSeparator *sep)
{
  sceneRoot = sep;
}


//-----------------------------------------------------------------------------
/**
 * Private attribute accessor.
 *
 * \return Whole scene root.
 *
 * \sa sceneRoot, setSceneRoot()
 */
SoSeparator* TeEngine::getSceneRoot()
{
  return sceneRoot;
}


//-----------------------------------------------------------------------------
/**
 * Adds \a node to scene.
 *
 * This function checks, whether the \a node is already present. If it is,
 * nothing happens. This prevents from adding the same \a node into the
 * scene several times. Helpful when adding seams into the scene graph.
 *
 * Note, that modifications to the scene graph are not allowed during the
 * rendering traversal of the tree. Violation of this condition results
 * usually in application crash.
 *
 * \param node Node to be added to the scene.
 *
 * \sa sceneRoot, removeNode()
 */
void TeEngine::addNode(SoNode *node)
{
  if (sceneRoot->findChild(node) == -1) sceneRoot->addChild(node);
}


//-----------------------------------------------------------------------------
/**
 * Removes \a node from scene.
 *
 * This function checks, whether the \a node is truly present. If it isn't,
 * nothing happens.
 *
 * Note, that modifications to the scene graph are not allowed during the
 * rendering traversal of the tree. Violation of this condition results
 * usually in application crash.
 *
 * \param node Node to be removed from the scene.
 *
 * \sa sceneRoot, addNode(), removePatch()
 */
void TeEngine::removeNode(SoNode *node)
{
  if (sceneRoot->findChild(node) != -1) sceneRoot->removeChild(node);
}


//-----------------------------------------------------------------------------
/**
 * Removes patch \a p and its seams from the scene.
 *
 * \param p Pointer to the patch; its graph is going to be removed from the
 *          scene.
 *
 * \sa removeNode()
 */
void TeEngine::removePatch(TePatch *p)
{
  if (p) { // this should happen if camera is moving too fast and we want
           // to remove patch, that is not generated  yet (it is scheduled)
    if (p->graph) removeNode(p->graph);
    for (int i=0; i<8; i++) {
      if (p->seamGraph[i]) removeNode(p->seamGraph[i]);
    }
  }
}


//-----------------------------------------------------------------------------
/**
 * Sets HUD scene root.
 *
 * \param sep Selected HUD scene root.
 *
 * \sa HUDRoot, getHUDRoot()
 */
void TeEngine::setHUDRoot(SoSeparator *sep)
{
  HUDRoot = sep;
}


/**
 * Private attribute accessor.
 *
 * \return HUD scene root.
 *
 * \sa HUDRoot, setHUDRoot()
 */
SoSeparator *TeEngine::getHUDRoot()
{
  return HUDRoot;
}


//-----------------------------------------------------------------------------
/**
 * Sets camera to be used to observe the scene.
 *
 * \param c Camera.
 *
 * \sa sceneCamera, getSceneCamera()
 */
void TeEngine::setSceneCamera(SoPerspectiveCamera *c)
{
  sceneCamera = c;
}


//-----------------------------------------------------------------------------
/**
 * Private attribute accessor.
 *
 * \return Camera used to observe the scene.
 *
 * \sa sceneCamera, setSceneCamera()
 */
SoPerspectiveCamera *TeEngine::getSceneCamera()
{
  return sceneCamera;
}


//-----------------------------------------------------------------------------
/**
 * Sets camera to be used to display the HUD.
 *
 * \param c Camera.
 *
 * \sa HUDCamera, getHUDCamera()
 */
void TeEngine::setHUDCamera(SoPerspectiveCamera *c)
{
  HUDCamera = c;
}


//-----------------------------------------------------------------------------
/**
 * Private attribute accessor.
 *
 * \return Camera used to display the HUD.
 *
 * \sa HUDCamera, setHUDCamera()
 */
SoPerspectiveCamera *TeEngine::getHUDCamera()
{
  return HUDCamera;
}


//-----------------------------------------------------------------------------
/**
 * Documentation not available.
 *
 * \todo More detailed documentation.
 */
TePatch* TeEngine::getPatch(const SbVec2f &pos, const int level, 
                            TePatch::PatchMissedPolicy policy)
{
  // handle root == NULL
  if (!root) {
    if (policy == TePatch::DONT_CREATE)
      return NULL;

    root = new TePatch(this);
    SbVec2f size2(patch0Size/2.f);
    int i;
    if (level > 0)
      for (i=0; i<level; i++)  size2 *= 2.f;
    else
      for (i=0; i>level; i--)  size2 *= 0.5f;

    root->makeRoot(SbBox2f(pos-size2, pos+size2), this);
    rootLevel = level;
    return root;
  }

  return root->getPatch(pos, level, policy);
}


//-----------------------------------------------------------------------------
/**
 * Private attribute accessor.
 *
 * \return Patch hierarchy root.
 *
 * \sa root
 */
TePatch* TeEngine::getRootPatch()
{
  return root;
}


//-----------------------------------------------------------------------------
/**
 * Private attribute accessor.
 *
 * \return Level of the patch hierarchy root.
 *
 * \sa rootLevel, getRootPatch()
 */
int TeEngine::getRootLevel() const
{
  return rootLevel;
}


//-----------------------------------------------------------------------------
/**
 * Documentation not available.
 *
 * \todo More detailed documentation.
 */
void TeEngine::advanceRootToPos(const SbVec2f &pos)
{
  if (!root) {
    root = new TePatch(this);
    SbVec2f size2(patch0Size/2.f);
    root->makeRoot(SbBox2f(pos-size2, pos+size2), this);
    rootLevel = 0;
    return;
  }


  while (!root->getArea().intersect(pos)) {

    const SbBox2f &rootArea = root->getArea();
    SbVec2f min = rootArea.getMin();
    SbVec2f max = rootArea.getMax();
    float sizex,sizey;
    rootArea.getSize(sizex,sizey);

    if (pos[0] < rootArea.getMin()[0])  min[0] -= sizex;
    else  max[0] += sizex;

    if (pos[1] < rootArea.getMin()[1])  min[1] -= sizey;
    else  max[1] += sizey;

    TePatch *newRoot = new TePatch(this);
    newRoot->makeRoot(SbBox2f(min,max), this);
    TePatch::Direction d = TePatch::getDirectionXFromVec(rootArea.getCenter() - newRoot->getArea().getCenter());
    assert(d != -1 && "This should never happen.");
    newRoot->setDetail(d,root);
    root = newRoot;
    rootLevel++;
  }
}


//----------------------------------------------------------------------------
/**
 * Generates new height map of the given resolution.
 *
 * The appearance of the result can be influenced by choosing the generation
 * and filtering method before calling this function. Note, that all
 * generator parameters (including the random seed) can be set in order
 * to customize the process of generating the result.
 *
 * \param res Desired resolution of the map.
 * \return Pointer to the new height map object.
 *
 * \sa setCurrGen(), setCurrFilter(), TeGenerator
 */
TeHeightMap* TeEngine::generateHMap(const SbVec2s &res)
{
  TeHeightMap *tmp = new TeHeightMap(res);
  TeGenerator *g = NULL;

  switch (currGen) {
    case FAULT_FORMATION: g = &faultFormation; break;
  }

  g->setHMap(tmp);
  g->generate();

  if (currFilter != NONE) {
    switch (currFilter) {
      case LINEAR: g = &linearFilter; break;
    }
    g->setHMap(tmp);
    g->generate();
  }

  return tmp;
}


//-----------------------------------------------------------------------------
/**
 * Generates new height map using fault-formation algorithm.
 *
 * This method is obsolete and should be removed. Use generateHMap() instead.
 *
 * \return Pointer to the new height map object.
 *
 * \todo This method is obsolete and should be removed.
 */
TeHeightMap* TeEngine::generateHMapFF()
{
  return this->generateHMapFF(this->getHMapResolution());
}


//-----------------------------------------------------------------------------
/**
 * Generates new height map using fault-formation algorithm.
 *
 * This method is obsolete and should be removed. Use generateHMap() instead.
 *
 * \param res Desired resolution of the map.
 * \return Pointer to the new height map object.
 *
 * \todo This method is obsolete and should be removed.
 */
TeHeightMap* TeEngine::generateHMapFF(const SbVec2s &res)
{
  return this->generateHMapFF(res, (unsigned int)time(NULL)+(rand()%100), 150, 0.05f, 2.0f);
}


//-----------------------------------------------------------------------------
/**
 * Generates new height map using fault-formation algorithm.
 *
 * This method is obsolete and should be removed. Use generateHMap() instead.
 *
 * \param res Desired resolution of the map.
 * \param seed Random seed that defines terrain shape.
 * \param num_faults Number of faults to be generated (number of iterations).
 * \param min_delta Minimal fault height.
 * \param max_delta Maximal fault height.
 * \return Pointer to the new height map object.
 *
 * \todo This method is obsolete and should be removed.
 */
TeHeightMap* TeEngine::generateHMapFF(const SbVec2s &res,
                                      const unsigned int seed, const int num_faults,
                                      const float min_delta, const float max_delta)
{
  TeHeightMap *hmap = new TeHeightMap(res);
  hmap->generate_FaultFormation(seed, num_faults, min_delta, max_delta);
  return hmap;
}


//-----------------------------------------------------------------------------
/**
 * Initializes the engine.
 *
 * Checks if all necessary attributes are set, generates initial patches and
 * creates HUD.
 *
 * If something is wrong terminates the execution and echoes the error.
 *
 * This function can be called only once and this should be done after adding
 * all desired objects and correctly setting the active one, after attaching
 * scene and HUD cameras. Repeated calls are simply ignored.
 */
void TeEngine::initialize()
{
  static SbBool initialized = FALSE;

  if (!initialized) {

    if ( !sceneRoot || !HUDRoot || !sceneCamera || !HUDCamera
         || !objectList.getLength() || !activeObject ) {
      SoDebugError::post("TeEngine::initialize",
                         "Some vital TeEngine's attribute is not set or there is no active object in the engine.\n\nThe application will close now.");
      exit(0);
    }

    // schedule creation of patches below inserted objects...
    for (int i=0; i<objectList.getLength(); i++)
      updateGraph(objectList[i]);

    // ...and force the generation
    while (taskQueue.getLength()) {
      if (this->genTick()) done();
      faultFormation.seed = (unsigned int)time(NULL)+(rand()%10000);
    }

    activeObject->createHUD();
    initialized = TRUE;
  }
}


//-----------------------------------------------------------------------------
/**
 * Time tick function of the engine.
 *
 * This function has to be called continuously by application that uses engine
 * in order to have chance to do its work. This function is responsible for
 * giving generators chance to work (it calls genTick() method) and checking
 * if all cameras have whole visible terain ready to render (it calls
 * updateGraph() for all engine objects and TeObject::timeTick() as well).
 *
 * \param delta Time elapsed since last call.
 */
void TeEngine::timeTick(const SbTime delta)
{
  // TEMPORARY solution, it will be better to pick 4 numbers when
  // scheduling the task
  faultFormation.seed = (unsigned int)time(NULL)+(rand()%10000);

  // fps
  static SbTime lastUpdate = SbTime::getTimeOfDay();
  SbTime current = SbTime::getTimeOfDay();
  if (current-lastUpdate >= 0.1) {
    fps = 1.0f/float(delta.getValue());
    lastUpdate = current;
  }

  // generator tick, should go first, because we don't want searching
  // in patch tree and first step of generation in one timeTick call
  if (taskQueue.getLength())
    if (this->genTick()) done();

  // update scenegraph to satisfy objects' needs, insert tasks to taskQueue
  // here, after genTick() call
  for (int i=0; i<objectList.getLength(); i++) {
    objectList[i]->timeTick(delta);
    updateGraph(objectList[i]);
  }
}


//-----------------------------------------------------------------------------
/**
 * Internal patch generator tick function.
 *
 * This method performs one TeGenerator::genStep() for the 1st task in the
 * queue. Process is controlled by the TeETask::TeETaskStatus values.
 *
 * \return TRUE if task finished and result is ready; done() should be called
 *         in this case.
 *
 * \todo Consider creating high-level "TePatchGenerator" that will do this
 *       work and make this function obsolete.
 */
SbBool TeEngine::genTick()
{
  TePatch *p = taskQueue[0].p;

  switch(taskQueue[0].status) {
      case TeETask::BM0: if (!p->getBuildMap(0)) {
                           p->findBuildMap(0);
                           if (!p->getBuildMap(0)) {
                             p->setBuildMap(0, new TeHeightMap(getBuildMapResolution()));
                             switch (taskQueue[0].gm) {
                               case FAULT_FORMATION: generator = &faultFormation; break;
                             }
                             generator->setHMap(p->getBuildMap(0));
                             taskQueue[0].status = TeETask::BM0G;
                           }
                           else taskQueue[0].status = TeETask::BM1;
                         }
                         else taskQueue[0].status = TeETask::BM1;
                         break;
     case TeETask::BM0G: if (generator->genStep()) {
                           taskQueue[0].status = TeETask::BM0F;
                           switch (taskQueue[0].fm) {
                               case NONE: taskQueue[0].status = TeETask::BM1; break;
                             case LINEAR: generator = &linearFilter; break;
                           }
                           generator->setHMap(p->getBuildMap(0));
                         }
                         break;
     case TeETask::BM0F: if (generator->genStep()) taskQueue[0].status = TeETask::BM0;
                         break;
      case TeETask::BM1: if (!p->getBuildMap(1)) {
                           p->findBuildMap(1);
                           if (!p->getBuildMap(1)) {
                             p->setBuildMap(1, new TeHeightMap(getBuildMapResolution()));
                             switch (taskQueue[0].gm) {
                               case FAULT_FORMATION: generator = &faultFormation; break;
                             }
                             generator->setHMap(p->getBuildMap(1));
                             taskQueue[0].status = TeETask::BM1G;
                           }
                           else taskQueue[0].status = TeETask::BM2;
                         }
                         else taskQueue[0].status = TeETask::BM2;
                         break;
     case TeETask::BM1G: if (generator->genStep()) {
                           taskQueue[0].status = TeETask::BM1F;
                           switch (taskQueue[0].fm) {
                               case NONE: taskQueue[0].status = TeETask::BM2; break;
                             case LINEAR: generator = &linearFilter; break;
                           }
                           generator->setHMap(p->getBuildMap(1));
                         }
                         break;
     case TeETask::BM1F: if (generator->genStep()) taskQueue[0].status = TeETask::BM1;
                         break;
      case TeETask::BM2: if (!p->getBuildMap(2)) {
                           p->findBuildMap(2);
                           if (!p->getBuildMap(2)) {
                             p->setBuildMap(2, new TeHeightMap(getBuildMapResolution()));
                             switch (taskQueue[0].gm) {
                               case FAULT_FORMATION: generator = &faultFormation; break;
                             }
                             generator->setHMap(p->getBuildMap(2));
                             taskQueue[0].status = TeETask::BM2G;
                           }
                           else taskQueue[0].status = TeETask::BM3;
                         }
                         else taskQueue[0].status = TeETask::BM3;
                         break;
     case TeETask::BM2G: if (generator->genStep()) {
                           taskQueue[0].status = TeETask::BM2F;
                           switch (taskQueue[0].fm) {
                               case NONE: taskQueue[0].status = TeETask::BM3; break;
                             case LINEAR: generator = &linearFilter; break;
                           }
                           generator->setHMap(p->getBuildMap(2));
                         }
                         break;
     case TeETask::BM2F: if (generator->genStep()) taskQueue[0].status = TeETask::BM2;
                         break;
      case TeETask::BM3: if (!p->getBuildMap(3)) {
                           p->findBuildMap(3);
                           if (!p->getBuildMap(3)) {
                             p->setBuildMap(3, new TeHeightMap(getBuildMapResolution()));
                             switch (taskQueue[0].gm) {
                               case FAULT_FORMATION: generator = &faultFormation; break;
                             }
                             generator->setHMap(p->getBuildMap(3));
                             taskQueue[0].status = TeETask::BM3G;
                           }
                           else taskQueue[0].status = TeETask::HM;
                         }
                         else taskQueue[0].status = TeETask::HM;
                         break;
     case TeETask::BM3G: if (generator->genStep()) {
                           taskQueue[0].status = TeETask::BM3F;
                           switch (taskQueue[0].fm) {
                               case NONE: taskQueue[0].status = TeETask::HM; break;
                             case LINEAR: generator = &linearFilter; break;
                           }
                           generator->setHMap(p->getBuildMap(3));
                         }
                         break;
     case TeETask::BM3F: if (generator->genStep()) taskQueue[0].status = TeETask::HM;
                         break;
       case TeETask::HM: if (!p->getHMap()) p->buildHMapFromBuildMaps();
                         taskQueue[0].status = TeETask::GRAPH;
                         break;
    case TeETask::GRAPH: p->getPatchGraph();
                         taskQueue[0].status = TeETask::SEAMS;
                         break;
    case TeETask::SEAMS: for (int i=0; i<8; i++) {
                           if (p->generateSeamMap(TePatch::Direction(i))) p->getSeamGraph(TePatch::Direction(i));
                         }
                         return TRUE;
                         break;
  }
  return FALSE;
}


//-----------------------------------------------------------------------------
/**
 * Tests whether the patch is already generated or not.
 *
 * Patch is specified by its center in the world-space coordinates passed
 * to this function in the \a patchPos argument. This is way how to
 * uniquely describe each patch.
 *
 * \param patchPos Position of the center of the patch in world-space.
 * \return TRUE if the specified patch is ready, FALSE if it isn't
 *
 * \todo Is the attribute TeEngine::patchReadyTree necessary? Try to use
 *       getPatch() with TePatch::DONT_CREATE policy instead.
 *
 * \sa patchReadyTree
 */
SbBool TeEngine::isReady(SbVec2f patchPos)
{
  return (patchReadyTree.findPoint(SbVec3f(patchPos[0],patchPos[1],0.0f)) != -1);
}


//-----------------------------------------------------------------------------
/**
 * Tests whether the patch is scheduled for generating or not.
 *
 * Patch is specified by its center in the world-space coordinates passed
 * to this function in the \a patchPos argument. This is way how to
 * uniquely describe each patch.
 *
 * \param patchPos Position of the center of the patch in world-space.
 * \return TRUE if the specified patch is scheduled, FALSE if it isn't
 *
 * \sa schedule(), patchScheduledTree
 */
SbBool TeEngine::isScheduled(SbVec2f patchPos)
{
  return (patchScheduledTree.findPoint(SbVec3f(patchPos[0],patchPos[1],0.0f)) != -1);
}


//-----------------------------------------------------------------------------
/**
 * Schedule generation of patch at \a patchPos.
 *
 * This function creates TeETask object, fills all its attributes and
 * inserts it into the TeEngine::taskQueue. To control the generation process,
 * modify TeEngine::currGen and TeEngine::currFilter attributes before
 * calling this method.
 *
 * \param patchPos Position of the center of the patch in world-space.
 *
 * \sa TeETask, setCurrGen(), setCurrFilter()
 */
void TeEngine::schedule(SbVec2f patchPos)
{
  TeETask task;
  task.p = getPatch(patchPos);
  task.patchPos = patchPos;
  task.status = TeETask::BM0;
  task.gm = currGen;
  task.fm = currFilter;

  taskQueue.append(task);
  tasksInQueue++;
  patchScheduledTree.addPoint(SbVec3f(patchPos[0],patchPos[1],0.0f));
}


//-----------------------------------------------------------------------------
/**
 * Removes task from the queue and marks the patch as ready.
 *
 * Called after succesfull completion of each task.
 *
 * \sa taskQueue
 */
void TeEngine::done()
{
  patchScheduledTree.removePoint(SbVec3f(taskQueue[0].patchPos[0],taskQueue[0].patchPos[1],0.0f));
  patchReadyTree.addPoint(SbVec3f(taskQueue[0].patchPos[0],taskQueue[0].patchPos[1],0.0f));
  taskQueue.remove(0);
  tasksInQueue--;
  patchesNum++;
}


//-----------------------------------------------------------------------------
/**
 * Schedules generation of all not existing neighbours of the patch at
 * \a patchPos.
 *
 * Patch at \a patchPos is likely to be under camera in a few moments, so look
 * in scheduled and ready patches and schedule generation of all necessary
 * neighbours.
 *
 * Little trick is used in this method: Count from 7 backwards is here because
 * NORTH neighbour is represented by nr.7 and with the camera currently
 * pointing to the north it is good to create north neighbour as the first
 * one.
 *
 * \param patchPos Position of the center of the patch in world-space.
 *
 * \todo Additional logic to assure that patches will be in the queue
 *       in optimal order (according to the movement speed vector). This will
 *       make the trick mentioned above obsolete.
 */
void TeEngine::prepareNeighbours(SbVec2f patchPos)
{
  for (int i=7; i>=0; i--) {
    if ( !isReady(computeNeighbourPos(patchPos, TePatch::Direction(i))) &&
         !isScheduled(computeNeighbourPos(patchPos, TePatch::Direction(i))) ) {
      schedule(computeNeighbourPos(patchPos, TePatch::Direction(i)));
    }
  }
}


//-----------------------------------------------------------------------------
/**
 * Updates scene graph to match object needs.
 *
 * This function is called continuously from the timeTick() method and
 * works in three steps:
 * \li Pre-generating: schedules patches that are going to be visible shortly.
 * \li Removes the patches that moved out of camera's field of view.
 * \li Checks if all visible terrain components (patches and seams) are
 *     in the scene graph.
 *
 * \param obj Object used for visibility checking.
 *
 * \todo Some optimization should be done here.
 */
void TeEngine::updateGraph(TeObject *obj)
{
  SbVec3f objPos = obj->position.getValue();
  SbVec2f patchPos = computePatchPosUnderPoint(objPos);
  TePatch::Direction objPosInside = computePosInsidePatch(patchPos, objPos);

  TePatch *p = getPatch(patchPos);
  int i = 0;

  // first call with this object -> prepare initial 3x3 landscape
  if (!obj->initialized) {
    if (!isReady(patchPos) && !isScheduled(patchPos)) schedule(patchPos);
    prepareNeighbours(patchPos);
    obj->initialized = TRUE;
    obj->patchPosUnder = patchPos;
    obj->posInsidePatch = objPosInside;
  }

  // 1) Pre-generating
  // object is moving, so pre-generate patches in the direction of movement
  // TODO: aditional logic to select the best according to the movement
  //       speed vector.
  //       - first compare old and current "posInsidePatch" and then
  //         make decision about what to schedule
  if ( (obj->patchPosUnder == patchPos)         // no new patch
       && (obj->posInsidePatch != objPosInside) // object pos changed
       && (objPosInside != -1) ) {              // not in the middle
    prepareNeighbours(computeNeighbourPos(patchPos, objPosInside));
  }

  // 2) Removing
  // - should go first, because later in this function we check if all
  //   visible things are shown and there we'll have a chance to undo
  //   incorrect changes made here.
  if ( obj->patchPosUnder != patchPos ) {
    switch (objPosInside) {
      case TePatch::SOUTH_WEST: switch (obj->posInsidePatch) {
                                        case TePatch::EAST:
                                  case TePatch::SOUTH_EAST: // same as WEST case below
                                                            removePatch(p->neighbour[TePatch::WEST]->neighbour[TePatch::NORTH_WEST]);
                                                            removePatch(p->neighbour[TePatch::WEST]->neighbour[TePatch::WEST]);
                                                            removePatch(p->neighbour[TePatch::WEST]->neighbour[TePatch::SOUTH_WEST]);
                                                            break;
                                  case TePatch::NORTH_EAST: // diagonal
                                                            removePatch(p->neighbour[TePatch::SOUTH_WEST]->neighbour[TePatch::NORTH_WEST]);
                                                            removePatch(p->neighbour[TePatch::SOUTH_WEST]->neighbour[TePatch::WEST]);
                                                            removePatch(p->neighbour[TePatch::SOUTH_WEST]->neighbour[TePatch::SOUTH_WEST]);
                                                            removePatch(p->neighbour[TePatch::SOUTH_WEST]->neighbour[TePatch::SOUTH]);
                                                            removePatch(p->neighbour[TePatch::SOUTH_WEST]->neighbour[TePatch::SOUTH_EAST]);
                                                            break;
                                  case TePatch::NORTH_WEST:
                                       case TePatch::NORTH: // same as SOUTH case below
                                                            removePatch(p->neighbour[TePatch::SOUTH]->neighbour[TePatch::SOUTH_WEST]);
                                                            removePatch(p->neighbour[TePatch::SOUTH]->neighbour[TePatch::SOUTH]);
                                                            removePatch(p->neighbour[TePatch::SOUTH]->neighbour[TePatch::SOUTH_EAST]);
                                                            break;
                                                   default: break; // this will never happen
                                }
                                break;
      case TePatch::SOUTH_EAST: switch (obj->posInsidePatch) {
                                        case TePatch::WEST:
                                  case TePatch::SOUTH_WEST: // same as EAST case below
                                                            removePatch(p->neighbour[TePatch::EAST]->neighbour[TePatch::NORTH_EAST]);
                                                            removePatch(p->neighbour[TePatch::EAST]->neighbour[TePatch::EAST]);
                                                            removePatch(p->neighbour[TePatch::EAST]->neighbour[TePatch::SOUTH_EAST]);
                                                            break;
                                  case TePatch::NORTH_WEST: // diagonal
                                                            removePatch(p->neighbour[TePatch::SOUTH_EAST]->neighbour[TePatch::NORTH_EAST]);
                                                            removePatch(p->neighbour[TePatch::SOUTH_EAST]->neighbour[TePatch::EAST]);
                                                            removePatch(p->neighbour[TePatch::SOUTH_EAST]->neighbour[TePatch::SOUTH_EAST]);
                                                            removePatch(p->neighbour[TePatch::SOUTH_EAST]->neighbour[TePatch::SOUTH]);
                                                            removePatch(p->neighbour[TePatch::SOUTH_EAST]->neighbour[TePatch::SOUTH_WEST]);
                                                            break;
                                  case TePatch::NORTH_EAST:
                                       case TePatch::NORTH: // same as SOUTH case below
                                                            removePatch(p->neighbour[TePatch::SOUTH]->neighbour[TePatch::SOUTH_WEST]);
                                                            removePatch(p->neighbour[TePatch::SOUTH]->neighbour[TePatch::SOUTH]);
                                                            removePatch(p->neighbour[TePatch::SOUTH]->neighbour[TePatch::SOUTH_EAST]);
                                                            break;
                                                   default: break; // this will never happen
                                }
                                break;
      case TePatch::NORTH_WEST: switch (obj->posInsidePatch) {
                                       case TePatch::SOUTH:
                                  case TePatch::SOUTH_WEST: // same as NORTH case below
                                                            removePatch(p->neighbour[TePatch::NORTH]->neighbour[TePatch::NORTH_WEST]);
                                                            removePatch(p->neighbour[TePatch::NORTH]->neighbour[TePatch::NORTH]);
                                                            removePatch(p->neighbour[TePatch::NORTH]->neighbour[TePatch::NORTH_EAST]);
                                                            break;
                                  case TePatch::SOUTH_EAST: // diagonal
                                                            removePatch(p->neighbour[TePatch::NORTH_WEST]->neighbour[TePatch::NORTH_EAST]);
                                                            removePatch(p->neighbour[TePatch::NORTH_WEST]->neighbour[TePatch::NORTH]);
                                                            removePatch(p->neighbour[TePatch::NORTH_WEST]->neighbour[TePatch::NORTH_WEST]);
                                                            removePatch(p->neighbour[TePatch::NORTH_WEST]->neighbour[TePatch::WEST]);
                                                            removePatch(p->neighbour[TePatch::NORTH_WEST]->neighbour[TePatch::SOUTH_WEST]);
                                                            break;
                                  case TePatch::NORTH_EAST:
                                        case TePatch::EAST: // same as WEST case below
                                                            removePatch(p->neighbour[TePatch::WEST]->neighbour[TePatch::NORTH_WEST]);
                                                            removePatch(p->neighbour[TePatch::WEST]->neighbour[TePatch::WEST]);
                                                            removePatch(p->neighbour[TePatch::WEST]->neighbour[TePatch::SOUTH_WEST]);
                                                            break;
                                                   default: break; // this will never happen
                                }
                                break;
      case TePatch::NORTH_EAST: switch (obj->posInsidePatch) {
                                       case TePatch::SOUTH:
                                  case TePatch::SOUTH_EAST: // same as NORTH case below
                                                            removePatch(p->neighbour[TePatch::NORTH]->neighbour[TePatch::NORTH_WEST]);
                                                            removePatch(p->neighbour[TePatch::NORTH]->neighbour[TePatch::NORTH]);
                                                            removePatch(p->neighbour[TePatch::NORTH]->neighbour[TePatch::NORTH_EAST]);
                                                            break;
                                  case TePatch::SOUTH_WEST: // diagonal - TODO
                                                            removePatch(p->neighbour[TePatch::NORTH_EAST]->neighbour[TePatch::NORTH_WEST]);
                                                            removePatch(p->neighbour[TePatch::NORTH_EAST]->neighbour[TePatch::NORTH]);
                                                            removePatch(p->neighbour[TePatch::NORTH_EAST]->neighbour[TePatch::NORTH_EAST]);
                                                            removePatch(p->neighbour[TePatch::NORTH_EAST]->neighbour[TePatch::EAST]);
                                                            removePatch(p->neighbour[TePatch::NORTH_EAST]->neighbour[TePatch::SOUTH_EAST]);
                                                            break;
                                  case TePatch::NORTH_WEST:
                                        case TePatch::WEST: // same as EAST case below
                                                            removePatch(p->neighbour[TePatch::EAST]->neighbour[TePatch::NORTH_EAST]);
                                                            removePatch(p->neighbour[TePatch::EAST]->neighbour[TePatch::EAST]);
                                                            removePatch(p->neighbour[TePatch::EAST]->neighbour[TePatch::SOUTH_EAST]);
                                                            break;
                                                   default: break; // this will never happen
                                }
                                break;
           case TePatch::SOUTH: removePatch(p->neighbour[TePatch::SOUTH]->neighbour[TePatch::SOUTH_WEST]);
                                removePatch(p->neighbour[TePatch::SOUTH]->neighbour[TePatch::SOUTH]);
                                removePatch(p->neighbour[TePatch::SOUTH]->neighbour[TePatch::SOUTH_EAST]);
                                break;
            case TePatch::WEST: removePatch(p->neighbour[TePatch::WEST]->neighbour[TePatch::NORTH_WEST]);
                                removePatch(p->neighbour[TePatch::WEST]->neighbour[TePatch::WEST]);
                                removePatch(p->neighbour[TePatch::WEST]->neighbour[TePatch::SOUTH_WEST]);
                                break;
            case TePatch::EAST: removePatch(p->neighbour[TePatch::EAST]->neighbour[TePatch::NORTH_EAST]);
                                removePatch(p->neighbour[TePatch::EAST]->neighbour[TePatch::EAST]);
                                removePatch(p->neighbour[TePatch::EAST]->neighbour[TePatch::SOUTH_EAST]);
                                break;
           case TePatch::NORTH: removePatch(p->neighbour[TePatch::NORTH]->neighbour[TePatch::NORTH_WEST]);
                                removePatch(p->neighbour[TePatch::NORTH]->neighbour[TePatch::NORTH]);
                                removePatch(p->neighbour[TePatch::NORTH]->neighbour[TePatch::NORTH_EAST]);
                                break;
                       default: break; // -1, this will never happen
    }
  }
  obj->posInsidePatch = objPosInside; // finally update info

  // 3) Adding
  // really simple :( way how to keep scenegraph updated
  // ---------------------------------------------------------------------
  // TODO: we are doing this every tick, this SHOULD be done differently
  //       although the performance seems to be quite good
  // ---------------------------------------------------------------------
  if (isReady(patchPos)) { // center patch and its seams
    addNode(p->graph);
    for (i=0; i<8; i++) {
      if (p->seamMap[i]) addNode(p->seamGraph[i]);
    }
  }
  if (obj == activeObject) { // only active obj 9x9, all others only 1 under
    for (i=0; i<8; i++) { // neighbours and theirs seams
      if (isReady(computeNeighbourPos(patchPos, TePatch::Direction(i)))) {
        addNode(p->neighbour[i]->graph);
        switch (i) {
           case 0: if (p->neighbour[i]->seamMap[TePatch::NORTH]) addNode(p->neighbour[i]->seamGraph[TePatch::NORTH]);
                   if (p->neighbour[i]->seamMap[TePatch::EAST]) addNode(p->neighbour[i]->seamGraph[TePatch::EAST]);
                   break;
           case 1: if (p->neighbour[i]->seamMap[TePatch::NORTH]) addNode(p->neighbour[i]->seamGraph[TePatch::NORTH]);
                   if (p->neighbour[i]->seamMap[TePatch::WEST]) addNode(p->neighbour[i]->seamGraph[TePatch::WEST]);
                   break;
           case 2: if (p->neighbour[i]->seamMap[TePatch::SOUTH]) addNode(p->neighbour[i]->seamGraph[TePatch::SOUTH]);
                   if (p->neighbour[i]->seamMap[TePatch::EAST]) addNode(p->neighbour[i]->seamGraph[TePatch::EAST]);
                   break;
           case 3: if (p->neighbour[i]->seamMap[TePatch::SOUTH]) addNode(p->neighbour[i]->seamGraph[TePatch::SOUTH]);
                   if (p->neighbour[i]->seamMap[TePatch::WEST]) addNode(p->neighbour[i]->seamGraph[TePatch::WEST]);
                   break;
          default: break;
        }
      }
    }
  }
  obj->patchPosUnder = patchPos; // finally update info
}


//-----------------------------------------------------------------------------
/**
 * Computes position of the patch that is right under the point \a pt.
 *
 * The position can be used to get the patch using getPatch() method.
 *
 * \param pt Point in 3D space.
 * \return Position of the patch under the point \a pt.
 *
 * \sa getPatch(), computeNeighbourPos()
 */
SbVec2f TeEngine::computePatchPosUnderPoint(SbVec3f pt)
{
  float px = patch0Size.getValue()[0];
  float py = patch0Size.getValue()[1];

  float divx = pt.getValue()[0]/px;
  float divy = pt.getValue()[1]/py;

  float x,y = 0;

  if (divx - floor(divx) > 0.5f) x = (float)(floor(divx)+1)*px;
  else x = (float)floor(divx)*px;

  if (divy - floor(divy) > 0.5f) y = (float)(floor(divy)+1)*py;
  else y = (float)floor(divy)*py;

  return SbVec2f(x, y);
}


//-----------------------------------------------------------------------------
/**
 * Computes position of the selected neighbour of the patch.
 *
 * The position can be used to get the neighbour using getPatch() method.
 *
 * \param patchPos Position of the center of the patch in world-space.
 * \param which Neighbour selector.
 * \return Position of the selected neighbour.
 *
 * \sa getPatch(), computePatchPosUnderPoint()
 */
SbVec2f TeEngine::computeNeighbourPos(SbVec2f patchPos,
                                      TePatch::Direction which)
{
  // if we put these directly down into switch, will it be faster??
  float px = patch0Size.getValue()[0];
  float py = patch0Size.getValue()[1];

  switch (which) {
    case TePatch::SOUTH_WEST: return patchPos + SbVec2f(-px, -py); break;
    case TePatch::SOUTH_EAST: return patchPos + SbVec2f( px, -py); break;
    case TePatch::NORTH_WEST: return patchPos + SbVec2f(-px,  py); break;
    case TePatch::NORTH_EAST: return patchPos + SbVec2f( px,  py); break;
         case TePatch::SOUTH: return patchPos + SbVec2f(  0, -py); break;
          case TePatch::WEST: return patchPos + SbVec2f(-px,   0); break;
          case TePatch::EAST: return patchPos + SbVec2f( px,   0); break;
         case TePatch::NORTH: return patchPos + SbVec2f(  0,  py); break;
                     default: return SbVec2f(0.0f, 0.0f); break; // to avoid warning, this will never happen
  }
}


//-----------------------------------------------------------------------------
/**
 * Computes position of the point \a pt above patch at \a patchPos relatively
 * to the patch borders.
 *
 * The position can be used to determine the direction the object is moving.
 *
 * \param patchPos Position of the center of the patch in world-space.
 * \param pt Point in 3D space.
 * \return TePatch::Direction, if above the central area, returns
 *         Direction(-1).
 */
TePatch::Direction TeEngine::computePosInsidePatch(SbVec2f patchPos,
                                                   SbVec3f pt)
{
  float px6 = patch0Size.getValue()[0] / 6.0f;
  float py6 = patch0Size.getValue()[1] / 6.0f;

  float northBorder = patchPos.getValue()[1] + py6;
  float southBorder = patchPos.getValue()[1] - py6;
  float  westBorder = patchPos.getValue()[0] - px6;
  float  eastBorder = patchPos.getValue()[0] + px6;

  float ptx = pt.getValue()[0];
  float pty = pt.getValue()[1];

  if (ptx < westBorder) {      // NW, W, SW
    if (pty < southBorder) return TePatch::SOUTH_WEST;
    else if (pty < northBorder) return TePatch::WEST;
    else return TePatch::NORTH_WEST;
  }
  else if (ptx < eastBorder) { // N, -1, S
    if (pty < southBorder) return TePatch::SOUTH;
    else if (pty < northBorder) return TePatch::Direction(-1);
    else return TePatch::NORTH;
  }
  else {                       // NE, E, SE
    if (pty < southBorder) return TePatch::SOUTH_EAST;
    else if (pty < northBorder) return TePatch::EAST;
    else return TePatch::NORTH_EAST;
  }
}
