//##################################################
//# $Id: TeHeightMap.cpp 20 2005-03-14 13:15:11Z peciva $

/*****************************************************************************\
 *
 * TeHeightMap.cpp
 *
 * TeHeightMap class implementation
 *
 * Author: Martin Havlek (xhavli15 AT stud.fit.vutbr.cz)
 * Contributors: PCJohn (peciva AT fit.vutbr.cz)
 *
 * ----------------------------------------------------------------------------
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, authors will kindly welcome
 * if you give them credit and keep their names with their source code.
 *
\*****************************************************************************/


#include "TeHeightMap.h"

#include <cstdlib>  // NULL, malloc, free, rand
#include <cstring>  // memcpy
#include <cmath>    // floor

//----------------------------------------------------------------------------
/**
 * Resets TeHeightMap::flags to zero, marks all stored statistics as invalid.
 *
 * \sa minValue, maxValue, averageValue
 */
void TeHeightMap::resetFlags()
{
  flags.minUpdated = flags.maxUpdated = flags.averageUpdated = 0;
}


//----------------------------------------------------------------------------
/**
 * Simple default constructor.
 *
 * Creates empty object (zero dimensions, no allocated memory).
 */
TeHeightMap::TeHeightMap() :
  resx(0),
  resy(0),
  values(NULL),
  minValue(0),
  maxValue(0),
  averageValue(0)
{
  resetFlags();
}


//----------------------------------------------------------------------------
/**
 * Constructor that sets dimensions and allocates memory.
 *
 * Creates initialized object with known dimensions and allocated memory
 * to store generated values.
 *
 * \param resolution Desired heightmap resolution.
 */
TeHeightMap::TeHeightMap(const SbVec2s resolution) :
  resx(resolution[0]),
  resy(resolution[1]),
  minValue(0),
  maxValue(0),
  averageValue(0)
{
  values = (float *)malloc(resx*resy*sizeof(float));
  resetFlags();
}


//----------------------------------------------------------------------------
/**
 * Constructor that sets dimensions and allocates memory.
 *
 * Creates initialized object with known dimensions and allocated memory
 * to store generated values.
 *
 * \param rx Desired resolution in X-axis direction.
 * \param ry Desired resolution in Y-axis direction.
 */
TeHeightMap::TeHeightMap(const int rx, const int ry) :
  resx(rx),
  resy(ry),
  minValue(0),
  maxValue(0),
  averageValue(0)
{
  values = (float *)malloc(resx*resy*sizeof(float));
  resetFlags();
}


//----------------------------------------------------------------------------
/**
 * Copy-constructor.
 *
 * Creates new heightmap as a copy of source heightmap.
 *
 * \param hmap Reference to source heightmap.
 */
TeHeightMap::TeHeightMap(const TeHeightMap &hmap) :
  resx(hmap.resx),
  resy(hmap.resy),
  minValue(hmap.minValue),
  maxValue(hmap.maxValue),
  averageValue(hmap.averageValue),
  flags(hmap.flags)
{
  values = (float *)malloc(resx*resy*sizeof(float));
  memcpy((void*)values, (const void*)hmap.values, resx*resy*sizeof(float));
}


//----------------------------------------------------------------------------
/**
 * Constructor that copies only specified part of source heightmap.
 *
 * New heightmap will have the same resolution as the source \a hmap.
 * Values are copied from the source \a hmap. The rectangular area of values
 * being copied is defined by \a origin and \a asize parameters. The values
 * are pasted at (0,0) coordinates.
 *
 * \param hmap Pointer to source heightmap.
 * \param origin Top-left corner of the copied area in the source heightmap.
 * \param asize Size of copied area.
 */
TeHeightMap::TeHeightMap(const TeHeightMap *hmap, const SbVec2s origin,
                         const SbVec2s asize) :
  resx(hmap->resx),
  resy(hmap->resy),
  minValue(0),
  maxValue(0),
  averageValue(0)
{
  values = (float *)malloc(resx*resy*sizeof(float));
  copyFrom(hmap, origin, asize);
}


//----------------------------------------------------------------------------
/**
 * Destructor.
 *
 * Releases memory.
 */
TeHeightMap::~TeHeightMap()
{
  free((void *)values);
}


//----------------------------------------------------------------------------
/**
 * Assignment operator
 *
 * \param hmap Reference to source heightmap.
 * \return Reference to itself.
 */
TeHeightMap& TeHeightMap::operator= (const TeHeightMap &hmap)
{
  if(&hmap != this) {
    resx = hmap.resx;
    resy = hmap.resy;
    if (values!=NULL) free((void *)values);
    values = (float *)malloc(resx*resy*sizeof(float));
    memcpy((void*)values, (const void*)hmap.values, resx*resy*sizeof(float));
    minValue = hmap.minValue;
    maxValue = hmap.maxValue;
    averageValue = hmap.averageValue;
    flags = hmap.flags;
  }
  return *this;
}


//----------------------------------------------------------------------------
/**
 * Returns size.
 * 
 * \return Vector representing resolution of heightmap.
 */
SbVec2s TeHeightMap::getResolution() const
{
  return SbVec2s(resx, resy);
}


//----------------------------------------------------------------------------
/**
 * Returns size.
 *
 * \param x Reference to the output variable for x-resolution.
 * \param y Reference to the output variable for y-resolution.
 */
void TeHeightMap::getResolution(int &x, int &y) const
{
  x = resx;
  y = resy;
}


//----------------------------------------------------------------------------
/**
 * Sets new resolution.
 *
 * If map was larger cut off previous values, if map was smaller set new
 * points to zero, old will be in the top-left corner.
 *
 * \param newRes Desired heightmap size.
 */
void TeHeightMap::setResolution(const SbVec2s &newRes)
{
  if (resx!=newRes[0] && resy != newRes[1]) {
    float *new_values = (float *)calloc(newRes[0]*newRes[1], sizeof(float));
    for (int i=0; i<((newRes[1]<resx) ? newRes[1] : resy); i++) {
      memcpy((void*)(new_values+i*newRes[0]), (const void*)(values+i*resx),
             (newRes[0]<resx) ? newRes[0]*sizeof(float)
                              : resx*sizeof(float));
    }
    free((void *)values);
    values = new_values;
    resx = newRes[0];
    resy = newRes[1];
    resetFlags();
  }
}


//----------------------------------------------------------------------------
/**
 * Sets one heightmap value.
 *
 * Sets value at (\a x, \a y) to \a value. If (\a x, \a y) is out of bounds
 * does nothing.
 *
 * \param x X coordinate of the point.
 * \param y Y coordinate of the point.
 * \param value Value to be placed at that point.
 */
void TeHeightMap::setValue(const int x, const int y, const float value)
{
  if ((x<resx) && (x>=0) && (y<resy) && (y>=0)) values[(y*resx)+x] = value;
  resetFlags();
}


//----------------------------------------------------------------------------
/**
 * Sets all heightmap values to specified \a value.
 *
 * \param value Float number to be set everywhere.
 *
 * \todo Consider implementing this using memset. Memset wants int argument,
 *       but for zero it should be fine too and it should be faster.
 */
void TeHeightMap::setAllValues(const float value)
{
  for (int i=0; i<resx*resy; i++) values[i] = value;
  flags.maxUpdated = flags.minUpdated = flags.averageUpdated = 1;
  maxValue = value;
  minValue = value;
  averageValue = value;
}


//----------------------------------------------------------------------------
/**
 * Returns value at specified coordinates.
 *
 * \param x X coordinate of the point.
 * \param y Y coordinate of the point.
 * \return Value at specified coordinates
 */
float TeHeightMap::getValue(const int x, const int y) const
{
  return values[(y*resx)+x];
}


//----------------------------------------------------------------------------
/**
 * Returns values array.
 *
 * \return Pointer to memory location, where heightmap values are stored.
 */
const float *TeHeightMap::getValues() const
{
  return values;
}


//----------------------------------------------------------------------------
/**
 * One row getting operator.
 *
 * Useful for coding like height=hmap[y][x].
 *
 * \param i Row index
 * \return Pointer to memory location of specified row.
 */
const float* TeHeightMap::operator[] (int i) const
{
  return &values[resx*i];
}


//----------------------------------------------------------------------------
/**
 * Starts editing.
 *
 * \return Pointer to memory location, where heightmap values are stored.
 */
float* TeHeightMap::startEditing()
{
  return values;
}


//----------------------------------------------------------------------------
/**
 * Finishes editing.
 *
 * It performs resetFlags().
 */
void TeHeightMap::finishEditing()
{
  resetFlags();
}


//----------------------------------------------------------------------------
/**
 * Statistics gathering (fast single-pass)
 *
 * Returns stats faster than when calling standalone routines, uses
 * TeHeightMap::flags so we can save even more time.
 *
 * If you do not need all the values, pass NULL as no-wanted args when
 * calling this function.
 *
 * \param min Pointer to the output variable for minimal value.
 * \param max Pointer to the output variable for maximal value.
 * \param average Pointer the to output variable for average value.
 *
 * \sa minValue, maxValue, averageValue, getMinValue(), getMaxValue(), getAverageValue()
 */
void TeHeightMap::getStats(float *min, float *max, float *average)
{
  if (min!=NULL || max!=NULL || average!=NULL) {
    if (flags.minUpdated && flags.maxUpdated && flags.averageUpdated) {
      if (min!=NULL) *min = minValue;
      if (max!=NULL) *max = maxValue;
      if (average!=NULL) *average = averageValue;
    }
    else {
      float sum = 0;
      minValue = 1000000;
      maxValue = -1000000;

      for (int i=0; i<resx*resy; i++) {
        sum += values[i];
        if (values[i]<minValue) minValue = values[i];
        if (values[i]>maxValue) maxValue = values[i];
      }

      averageValue = sum/(resx*resy);

      flags.minUpdated = 1;
      flags.maxUpdated = 1;
      flags.averageUpdated = 1;

      if (min!=NULL) *min = minValue;
      if (max!=NULL) *max = maxValue;
      if (average!=NULL) *average = averageValue;
    }
  }
}


//----------------------------------------------------------------------------
/**
 * Finds minimal value in entire heightmap.
 *
 * Uses TeHeightMap::flags for speedup.
 *
 * \return Minimal height from the heightmap values.
 *
 * \sa getStats(), minValue
 */
float TeHeightMap::getMinValue()
{
  if (!flags.minUpdated) {
    minValue = 1000000;
    for (int i=0; i<resx*resy; i++)
      if (values[i]<minValue) minValue = values[i];
    flags.minUpdated = 1;
  }
  return minValue;
}


//----------------------------------------------------------------------------
/**
 * Finds maximal value in entire heightmap.
 *
 * Uses TeHeightMap::flags for speedup.
 *
 * \return Maximal height from the heightmap values.
 *
 * \sa getStats(), maxValue
 */
float TeHeightMap::getMaxValue()
{
  if (!flags.maxUpdated) {
    maxValue = -1000000;
    for (int i=0; i<resx*resy; i++)
      if (values[i]>maxValue) maxValue = values[i];
    flags.maxUpdated = 1;
  }
  return maxValue;
}


//----------------------------------------------------------------------------
/**
 * Computes average value from entire heightmap.
 *
 * Uses TeHeightMap::flags for speedup.
 *
 * \return Average height from the heightmap values.
 *
 * \sa getStats(), averageValue
 */
float TeHeightMap::getAverageValue()
{
  if (!flags.averageUpdated) {
    float sum = 0;
    for (int i=0; i<resx*resy; i++)
      sum += values[i];
    averageValue = sum/(resx*resy);
    flags.averageUpdated = 1;
  }
  return averageValue;
}


//----------------------------------------------------------------------------
/**
 * Adds \a delta to all values.
 *
 * \param delta This value will be added to all heightmap values.
 */
void TeHeightMap::shift(const float delta)
{
  if (delta != 0) {
    for (int i=0; i<resx*resy; i++) {
      values[i] += delta;
    }
    minValue += delta;
    maxValue += delta;
    averageValue += delta; // ?? OK ??
  }
}


//----------------------------------------------------------------------------
/**
 * Shifts heightmap values, \a value will become its minimal value.
 *
 * \param value Value that will become new minimal value in the heightmap.
 */
void TeHeightMap::shiftMinValue(const float value)
{
  shift(value - getMinValue());
}


//----------------------------------------------------------------------------
/**
 * Shifts heightmap values, \a value will become its minimal value.
 *
 * We know \a old minimal value, so we do not have to compute it (it's faster,
 * no check).
 *
 * \param value Value that will become new minimal value in the heightmap.
 * \param old Old minimal value (no check).
 */
void TeHeightMap::shiftMinValue(const float value, const float old)
{
  shift(value - old);
}


//----------------------------------------------------------------------------
/**
 * Shifts heightmap values, \a value will become its maximal value.
 *
 * \param value Value that will become new maximal value in the heightmap.
 */
void TeHeightMap::shiftMaxValue(const float value)
{
  shift(value - getMaxValue());
}


//----------------------------------------------------------------------------
/**
 * Shifts heightmap values, \a value will become its maximal value.
 *
 * We know \a old maximal value, so we do not have to compute it (it's faster,
 * no check).
 *
 * \param value Value that will become new maximal value in the heightmap.
 * \param old Old maximal value (no check).
 */
void TeHeightMap::shiftMaxValue(const float value, const float old)
{
  shift(value - old);
}


//----------------------------------------------------------------------------
/**
 * Shifts heightmap values, \a value will become its average value.
 *
 * \param value Value that will become new average value in the heightmap.
 */
void TeHeightMap::shiftAverageValue(const float value)
{
  shift(value - getAverageValue());
}


//----------------------------------------------------------------------------
/**
 * Shifts heightmap values, \a value will become its average value.
 *
 * We know \a old average value, so we do not have to compute it (it's faster,
 * no check).
 *
 * \param value Value that will become new average value in the heightmap.
 * \param old Old average value (no check).
 */
void TeHeightMap::shiftAverageValue(const float value, const float old)
{
  shift(value - old);
}


//----------------------------------------------------------------------------
/**
 * Copies all possible values from source heightmap.
 *
 * Does NOT CHANGE SIZE of the heightmap. If size of heightmaps differs, only
 * subset is copied.
 *
 * \param hmap Pointer to the source heightmap.
 */
void TeHeightMap::copyFrom(const TeHeightMap *hmap)
{
  if (resx==hmap->resx && resy==hmap->resy)
    this->operator=(*hmap);
  else {
    copyFrom(hmap, SbVec2s(0,0), getResolution());
    resetFlags();
  }
}


//----------------------------------------------------------------------------
/**
 * Copies selected part of source heightmap.
 *
 * Copied area is specified by its top-left corner \a sorigin and \a size,
 * values are pasted at \a dorigin coordinates.
 *
 * \param hmap Pointer to the source heightmap.
 * \param sorigin Top-left corner of selected area in source heightmap.
 * \param size Size of selected area.
 * \param dorigin Top-left corner of area in target heightmap.
 */
void TeHeightMap::copyFrom(const TeHeightMap *hmap, const SbVec2s &sorigin,
                           const SbVec2s &size, const SbVec2s &dorigin)
{
  if (sorigin[0]<hmap->resx && sorigin[1]<hmap->resy && resx>0 && resy>0) {

    // dimension-check in source
    int sizex = ((sorigin[0]+size[0]<=hmap->resx) ? size[0] : hmap->resx-sorigin[0] );
    int sizey = ((sorigin[1]+size[1]<=hmap->resy) ? size[1] : hmap->resy-sorigin[1] );

    // dimension-check in target
    if (sizex>this->resx-dorigin[0]) sizex = this->resx-dorigin[0];
    if (sizey>this->resy-dorigin[1]) sizey = this->resy-dorigin[1];

    if (sizex>0 && sizey>0) {
      // copying
      for (int i=0; i<sizey; i++) {
        memcpy( (void*)(values + ((dorigin[1]*this->resx)+(i*this->resx)+dorigin[0])),
                (const void*)(hmap->values + ((sorigin[1]*hmap->resx)+(i*hmap->resx)+sorigin[0])),
                sizex*sizeof(float) );
      }
      resetFlags();
    }
  }
}


//----------------------------------------------------------------------------
/**
 * Two heightmaps all-values adding.
 *
 * Values from source heightmap will be added to "this" heightmap according
 * to a "blend-rate" \a rate. This should be greater than 0.0f. All source
 * values are multiplied by \a rate before adding to target values.
 *
 * All possible values will be added, both source and target top-left
 * corners are in (0,0) coordinates.
 *
 * \param hmap Pointer to the source heightmap.
 * \param rate Blend-rate showing how much of source heightmap to be added
 *             to the target.
 */
void TeHeightMap::addFrom(const TeHeightMap *hmap, const float rate)
{
  addFrom(hmap, SbVec2s(0,0), getResolution(), SbVec2s(0,0), rate);
}


//----------------------------------------------------------------------------
/**
 * Two heightmaps specified-values adding.
 *
 * Values from source heightmap will be added to "this" heightmap according
 * to a "blend-rate" \a rate. This should be greater than 0.0f. All source
 * values are multiplied by \a rate before adding to target values.
 *
 * Selected source area is specified by its top-left corner \a sorogin and its
 * \a size. Target area is specified by its top-left corner \a dorogin.
 *
 * \param hmap Pointer to the source heightmap.
 * \param sorigin Top-left corner of selected area in source heightmap.
 * \param size Size of selected area.
 * \param dorigin Top-left corner of area in target heightmap.
 * \param rate Blend-rate showing how much of source heightmap to be added
 *             to the target.
 */
void TeHeightMap::addFrom(const TeHeightMap *hmap, const SbVec2s &sorigin,
                          const SbVec2s &size, const SbVec2s &dorigin,
                          const float rate)
{
  if (sorigin[0]<hmap->resx && sorigin[1]<hmap->resy && resx>0 && resy>0 && rate>0) {

    // dimension-check in source
    int sizex = ((sorigin[0]+size[0]<=hmap->resx) ? size[0] : hmap->resx-sorigin[0] );
    int sizey = ((sorigin[1]+size[1]<=hmap->resy) ? size[1] : hmap->resy-sorigin[1] );

    // dimension-check in target
    if (sizex>this->resx-dorigin[0]) sizex = this->resx-dorigin[0];
    if (sizey>this->resy-dorigin[1]) sizey = this->resy-dorigin[1];

    if (sizex>0 && sizey>0) {
      // adding
      for (int y=0; y<sizey; y++) {
        for (int x=0; x<sizex; x++) {
          values[(y*this->resx)+(dorigin[1]*this->resx)+dorigin[0]+x] +=
            hmap->values[(y*hmap->resx)+(sorigin[1]*hmap->resx)+sorigin[0]+x]*rate;
        }
      }
      resetFlags();
    }
  }
}


//----------------------------------------------------------------------------
/**
 * Adds constant to all heightmap values.
 *
 * Same as shift(value).
 *
 * \param value Constant to be added to all heightmap values.
 *
 * \sa shift()
 */
void TeHeightMap::addAll(const float value)
{
  shift(value);
}


//----------------------------------------------------------------------------
/**
 * Two heightmaps all-values subtracting.
 *
 * Values from source heightmap will be subtracted from "this" heightmap
 * according to a "blend-rate" \a rate. This should be greater than 0.0f. All
 * source values are multiplied by \a rate before subtracting from target
 * values.
 *
 * All possible values will be subtracted, both source and target top-left
 * corners are in (0,0) coordinates.
 *
 * \param hmap Pointer to the source heightmap.
 * \param rate Blend-rate showing how much of source heightmap to be
 *             subtracted from the target.
 */
void TeHeightMap::subtractFrom(const TeHeightMap *hmap, const float rate)
{
  subtractFrom(hmap, SbVec2s(0,0), getResolution(), SbVec2s(0,0), rate);
}


//----------------------------------------------------------------------------
/**
 * Two heightmaps specified-values subtracting.
 *
 * Values from source heightmap will be subtracted from "this" heightmap
 * according to a "blend-rate" \a rate. This should be greater than 0.0f. All
 * source values are multiplied by \a rate before subtracting from target
 * values.
 *
 * Selected source area is specified by its top-left corner \a sorogin and its
 * \a size. Target area is specified by its top-left corner \a dorogin.
 *
 * \param hmap Pointer to the source heightmap.
 * \param sorigin Top-left corner of selected area in source heightmap.
 * \param size Size of selected area.
 * \param dorigin Top-left corner of area in target heightmap.
 * \param rate Blend-rate showing how much of source heightmap to be subtracted
 *             from the target.
 */
void TeHeightMap::subtractFrom(const TeHeightMap *hmap, const SbVec2s &sorigin,
                               const SbVec2s &size, const SbVec2s &dorigin,
                               const float rate)
{
  if (sorigin[0]<hmap->resx && sorigin[1]<hmap->resy && size[0]>0 && size[1]>0 && rate>0) {

    // dimension-check in source
    int sizex = ((sorigin[0]+size[0]<=hmap->resx) ? size[0] : hmap->resx-sorigin[0] );
    int sizey = ((sorigin[1]+size[1]<=hmap->resy) ? size[1] : hmap->resy-sorigin[1] );
    
    // dimension-check in target
    if (sizex>this->resx-dorigin[0]) sizex = this->resx-dorigin[0];
    if (sizey>this->resy-dorigin[1]) sizey = this->resy-dorigin[1];

    if (sizex>0 && sizey>0) {
      // adding
      for (int y=0; y<sizey; y++) {
        for (int x=0; x<sizex; x++) {
          values[(y*this->resx)+(dorigin[1]*this->resx)+dorigin[0]+x] -=
            hmap->values[(y*hmap->resx)+(sorigin[1]*hmap->resx)+sorigin[0]+x]*rate;
        }
      }
      resetFlags();
    }
  }
}


//----------------------------------------------------------------------------
/**
 * Subtracts constant from all heightmap values.
 *
 * Same as shift(-value).
 *
 * \param value Constant to be subtracted from all heightmap values.
 *
 * \sa shift()
 */
void TeHeightMap::subtractAll(const float value)
{
  shift(-value);
}


//----------------------------------------------------------------------------
/**
 * Two heightmaps all-values multiplying.
 *
 * Values in target heightmap will be multiplied by values in the source
 * heightmap according to a "blend-rate" \a rate. This should be greater than
 * 0.0f. All source values are multiplied by \a rate before multiplying the
 * target values.
 *
 * All possible values will be multiplied, both source and target top-left
 * corners are in (0,0) coordinates.
 *
 * \param hmap Pointer to the source heightmap.
 * \param rate Blend-rate showing how much of source heightmap to be
 *             multiplied with.
 */
void TeHeightMap::multiplyFrom(const TeHeightMap *hmap, const float rate)
{
  multiplyFrom(hmap, SbVec2s(0,0), getResolution(), SbVec2s(0,0), rate);
}


//----------------------------------------------------------------------------
/**
 * Two heightmaps specified-values multiplying.
 *
 * Values in target heightmap will be multiplied by values in the source
 * heightmap according to a "blend-rate" \a rate. This should be greater than
 * 0.0f. All source values are multiplied by \a rate before multiplying the
 * target values.
 *
 * Selected source area is specified by its top-left corner \a sorogin and its
 * \a size. Target area is specified by its top-left corner \a dorogin.
 *
 * \param hmap Pointer to the source heightmap.
 * \param sorigin Top-left corner of selected area in source heightmap.
 * \param size Size of selected area.
 * \param dorigin Top-left corner of area in target heightmap.
 * \param rate Blend-rate showing how much of source heightmap to be
 *             multiplied with.
 */
void TeHeightMap::multiplyFrom(const TeHeightMap *hmap, const SbVec2s &sorigin,
                               const SbVec2s &size, const SbVec2s &dorigin,
                               const float rate)
{
  if (sorigin[0]<hmap->resx && sorigin[1]<hmap->resy && size[0]>0 && size[1]>0 && rate>0) {

    // dimension-check in source
    int sizex = ((sorigin[0]+size[0]<=hmap->resx) ? size[0] : hmap->resx-sorigin[0] );
    int sizey = ((sorigin[1]+size[1]<=hmap->resy) ? size[1] : hmap->resy-sorigin[1] );

    // dimension-check in target
    if (sizex>this->resx-dorigin[0]) sizex = this->resx-dorigin[0];
    if (sizey>this->resy-dorigin[1]) sizey = this->resy-dorigin[1];

    if (sizex>0 && sizey>0) {
      // adding
      for (int y=0; y<sizey; y++) {
        for (int x=0; x<sizex; x++) {
          values[(y*this->resx)+(dorigin[1]*this->resx)+dorigin[0]+x] *=
            hmap->values[(y*hmap->resx)+(sorigin[1]*hmap->resx)+sorigin[0]+x]*rate;
        }
      }
      resetFlags();
    }
  }
}


//----------------------------------------------------------------------------
/**
 * Multiplies all heightmap values by a constant.
 *
 * \param value All heightmap values will be multiplied by this constant.
 */
void TeHeightMap::multiplyAll(const float value)
{
  for (int i=0; i<resx*resy; i++)
    values[i] *= value;
  resetFlags();
}


//----------------------------------------------------------------------------
/**
 * Linear filtering of heightmap values
 *
 * In one cycle, the filter goes through the data in all rows and columns
 * in both directions and changes the values according to the neighbours and
 * filtration coefficient \a coef. This value should be between
 * (0.0f-1.0f). When large - smaller changes, when small - huge filtering will
 * be done. This cycle will be done \a times -times.
 *
 * \param coef Filtration coefficient, when large - little change, when small -
               huge filtering.
 * \param times Filtering will be done times -times
 *
 * \sa TeLinearFilter
 */
void TeHeightMap::filter_Linear(const float coef, const int times)
{
  if (coef<1.0 && coef>0.0) {
    int x = 0;
    int y = 0;
    for (int i=0; i<times; i++) {

      // rows
      for (y=0; y<resy; y++) {
        // from left to right
        x = 1;
        while (x < resx-1) {
          values[(y*resx)+x] = values[(y*resx)+x-1]*(1-coef)+values[(y*resx)+x]*coef;
          x++;
        }
        // from right to left
        x = resx-2;
        while (x > -1) {
          values[(y*resx)+x] = values[(y*resx)+x+1]*(1-coef)+values[(y*resx)+x]*coef;
          x--;
        }
      }
      // columns
      for(x=0; x<resx; x++) {
        // columns top to bottom
        y = 1;
        while (y < resy) {
          values[(y*resx)+x] = values[((y-1)*resx)+x]*(1-coef)+values[(y*resx)+x]*coef;
          y++;
        }
        // columns bottom to top
        y = resy-2;
        while (y > -1) {
          values[(y*resx)+x] = values[((y+1)*resx)+x]*(1-coef)+values[(y*resx)+x]*coef;
          y--;
        }
      }

    } // for
    resetFlags();
  } // if
}


//----------------------------------------------------------------------------
/**
 * Generates mask that can used when building heightmap from 4 others.
 *
 * Heightmaps have to be squares sized (2^x)+1 !!!
 *
 * \todo This method is obsolete and should be removed.
 */
void TeHeightMap::generate_4HMOverlap_Mask()
{
  int size = resx; // square !!!

  float step = (float)(1.0/floor(size/2));
  int base = (int)(floor(size/2));
  int center = base*size + base;

  int idx1 = center;
  int idx2 = idx1;
  int idx3 = idx1;
  int idx4 = idx1;

  int i, j = 0;
  float value;

  // init
  values[idx1] = 1.0;

  // center cross
  for (i=1; i<base; i++) {
    idx1--;
    idx2 -= size;
    idx3++;
    idx4 += size;
    value = (float)(1.0 - i*step);
    values[idx1] = values[idx2] = values[idx3] = values[idx4] = value;
  }

  // remaining values (the same for 4 quadrants do it "in parallel")
  for (i=1; i<base; i++) {
    for (j=1; j<base; j++) {
      idx1 = center - i*size - j;
      idx2 = center - i*size + j;
      idx3 = center + i*size - j;
      idx4 = center + i*size + j;
      value = values[center+j]*values[center+i*size];
      values[idx1] = values[idx2] = values[idx3] = values[idx4] = value;
    }
  }
}


//----------------------------------------------------------------------------
/**
 * Generates heightmap values using fault-formation algorithm.
 *
 * The algorithm works as follows:
 * \li Two points in the heightmap are selected. These points define the fault
 *     vector. The selection is done randomly (the generator is initialized
 *     with \a seed as random seed).
 * \li According to current iteration number, \a min_delta and
 *     \a max_delta, the height of the fault is computed.
 * \li All values in one half-plane specified by the fault are increased
 *     by computed value. Rest of values is left unchanged.
 * \li This cycle is done \a num_faults -times.
 *
 * \param seed Random seed that defines terrain shape.
 * \param num_faults Number of faults to be generated (number of iterations).
 * \param min_delta Minimal fault height.
 * \param max_delta Maximal fault height.
 *
 * \sa TeFaultFormation
 */
void TeHeightMap::generate_FaultFormation(const unsigned int seed,
                                          const int num_faults,
                                          const float min_delta,
                                          const float max_delta)
{
  // init
  srand(seed);
  setAllValues(0);

  float Delta = 0;

  unsigned int Rand1X, Rand1Y, Rand2X, Rand2Y;
  int FaultVectorX, FaultVectorY;
  int ToPointVectorX, ToPointVectorY;

  for (int i = 0; i < num_faults; i++) {
    Rand1X = rand() % resx;
    Rand1Y = rand() % resy;
    Rand2X = rand() % resx;
    Rand2Y = rand() % resy;

    // fault vector
    FaultVectorX = Rand2X - Rand1X;
    FaultVectorY = Rand2Y - Rand1Y;

    // value to be added to one side of the fault
    Delta = static_cast<float>
            (max_delta - ((max_delta-min_delta) * i) / num_faults);

    int idx=0;

    for (int y = 0; y < resy; y++) {
      for (int x = 0; x < resx; x++) {
        // vector to the current map position
        ToPointVectorX = x - Rand1X;
        ToPointVectorY = y - Rand1Y;
        if ((ToPointVectorX*FaultVectorY-FaultVectorX*ToPointVectorY) > 0)
          values[idx] += Delta;
        idx++;
      }
    }
  }
  resetFlags();
}


//----------------------------------------------------------------------------
// FOR DEBUGGING ONLY
//----------------------------------------------------------------------------

#include <cstdio>

/**
 * Dumps itself to stdout.
 *
 * This method is meant to be for DEBUGGING only and can be removed at any
 * time.
 */
void TeHeightMap::dump()
{
  printf("\n\n----- TeHeightMap::dump() ---------------------------------\n\n");

  printf(" Width: %d\n", resx);
  printf("Height: %d\n\n", resy);
  
  printf("Memory: %d B\n\n", sizeof(TeHeightMap)+(resx*resy*sizeof(float)));

  printf("    minValue: %6.2f (flag: %d)\n", minValue, flags.minUpdated);
  printf("    maxValue: %6.2f (flag: %d)\n", maxValue, flags.maxUpdated);
  printf("averageValue: %6.2f (flag: %d)\n\n", averageValue, flags.averageUpdated);

  if (values!=NULL) {
    printf(" Values: ");
    for (int i=0; i<resx*resy; i++) {
      if ((i%resx)==0 && (i!=0)) printf("\n         ");
      printf("%6.2f, ", values[i]);
    }
  }
  else printf(" Values: NULL");

  printf("\n\n-----------------------------------------------------------\n\n");
}
